"""Environment variable registry loading and management."""

from __future__ import annotations

import os
from pathlib import Path
from typing import Any

import yaml

from nim_audit.models.env import (
    Affect,
    ImpactLevel,
    ImpactMetric,
    InteractionEdge,
    Registry,
    RegistryEntry,
)

CONTROLLED_METRICS = ["latency", "throughput", "memory", "determinism", "numerics", "compatibility"]
CONTROLLED_IMPACTS = ["++", "+", "±", "-", "--", "none"]


def get_default_registry_path() -> str | None:
    """Get the default registry path from the package data directory."""
    data_dir = Path(__file__).parent.parent.parent / "data"
    registry_path = data_dir / "env_registry.v1.yaml"
    if registry_path.exists():
        return str(registry_path)
    return None


def get_default_interactions_path() -> str | None:
    """Get the default interactions path from the package data directory."""
    data_dir = Path(__file__).parent.parent.parent / "data"
    interactions_path = data_dir / "interactions.v1.yaml"
    if interactions_path.exists():
        return str(interactions_path)
    return None


def _parse_affect_item(it: Any) -> Affect | None:
    """Parse an affect item from various formats.

    Supported formats:
    - String: "latency:+"
    - Dict with metric/impact keys: {metric: "latency", impact: "+"}
    - Dict with metric as key: {latency: "+"}
    """
    if isinstance(it, str) and ":" in it:
        m, imp = it.split(":", 1)
        metric_str = str(m).strip().lower()
        impact_str = str(imp).strip()
        try:
            metric = ImpactMetric(metric_str)
            impact = ImpactLevel(impact_str)
            return Affect(metric=metric, impact=impact)
        except ValueError:
            return Affect(
                metric=ImpactMetric(metric_str) if metric_str in [e.value for e in ImpactMetric] else ImpactMetric.LATENCY,
                impact=ImpactLevel(impact_str) if impact_str in [e.value for e in ImpactLevel] else ImpactLevel.NONE,
            )
    if isinstance(it, dict):
        # Format: {metric: "latency", impact: "+"}
        if "metric" in it and "impact" in it:
            metric_str = str(it["metric"]).lower()
            impact_str = str(it["impact"])
            try:
                metric = ImpactMetric(metric_str)
                impact = ImpactLevel(impact_str)
                return Affect(metric=metric, impact=impact)
            except ValueError:
                return None
        # Format: {latency: "+"} - single key-value where key is metric
        if len(it) == 1:
            for metric_key, impact_val in it.items():
                metric_str = str(metric_key).lower()
                impact_str = str(impact_val)
                try:
                    metric = ImpactMetric(metric_str)
                    impact = ImpactLevel(impact_str)
                    return Affect(metric=metric, impact=impact)
                except ValueError:
                    return None
    return None


def _normalize_affects(raw: Any, warnings: list[str], var: str) -> list[Affect]:
    """Normalize affects list from raw YAML data."""
    out: list[Affect] = []
    if raw is None:
        return out
    if not isinstance(raw, list):
        warnings.append(f"{var}: affects should be a list")
        return out
    for it in raw:
        a = _parse_affect_item(it)
        if not a:
            warnings.append(f"{var}: invalid affects item {it!r}")
            continue
        if a.metric.value not in CONTROLLED_METRICS:
            warnings.append(f"{var}: affects.metric '{a.metric.value}' not in {CONTROLLED_METRICS}")
        if a.impact.value not in CONTROLLED_IMPACTS:
            warnings.append(f"{var}: affects.impact '{a.impact.value}' not in {CONTROLLED_IMPACTS}")
        out.append(a)
    # Deduplicate
    seen: set[tuple[str, str]] = set()
    unique: list[Affect] = []
    for a in out:
        key = (a.metric.value, a.impact.value)
        if key not in seen:
            seen.add(key)
            unique.append(a)
    return unique


def load_registry(
    registry_path: str | None = None,
    interactions_path: str | None = None,
) -> Registry:
    """
    Load the environment variable registry from YAML files.

    Args:
        registry_path: Path to registry YAML file (uses default if None)
        interactions_path: Path to interactions YAML file (uses default if None)

    Returns:
        Registry with entries and interactions
    """
    warnings: list[str] = []
    entries: dict[str, RegistryEntry] = {}
    edges: list[InteractionEdge] = []

    # Use defaults if not provided
    if registry_path is None:
        registry_path = get_default_registry_path()
    if interactions_path is None:
        interactions_path = get_default_interactions_path()

    raw: dict[str, Any] = {}
    if registry_path and os.path.exists(registry_path):
        with open(registry_path, "r", encoding="utf-8") as f:
            raw = yaml.safe_load(f) or {}

    items = raw.get("vars") if isinstance(raw, dict) else raw
    if items is None:
        items = []
    if not isinstance(items, list):
        warnings.append("registry: expected top-level 'vars' list")
        items = []

    for it in items:
        if not isinstance(it, dict) or "name" not in it:
            continue
        name = str(it["name"]).strip()
        affects = _normalize_affects(it.get("affects"), warnings, name)
        confidence = str(it.get("confidence") or "LOW").upper()
        if confidence not in ("HIGH", "MED", "LOW"):
            warnings.append(f"{name}: confidence '{confidence}' not in HIGH|MED|LOW")
            confidence = "LOW"
        entries[name] = RegistryEntry(
            name=name,
            type=str(it.get("type")) if it.get("type") is not None else None,
            scope=str(it.get("scope")) if it.get("scope") is not None else None,
            precedence=str(it.get("precedence")) if it.get("precedence") is not None else None,
            default=str(it.get("default")) if it.get("default") is not None else None,
            affects=affects,
            determinism=str(it.get("determinism")) if it.get("determinism") is not None else None,
            interactions=list(it.get("interactions") or []),
            failure_modes=[str(x) for x in (it.get("failure_modes") or [])],
            confidence=confidence,
            evidence=list(it.get("evidence") or []),
        )

    if interactions_path and os.path.exists(interactions_path):
        with open(interactions_path, "r", encoding="utf-8") as f:
            iraw = yaml.safe_load(f) or {}
        ed_items = iraw.get("edges") if isinstance(iraw, dict) else iraw
        if ed_items is None:
            ed_items = []
        if isinstance(ed_items, list):
            for e in ed_items:
                if not isinstance(e, dict):
                    continue
                if not all(k in e for k in ("var_a", "var_b", "interaction_type", "description")):
                    continue
                edges.append(
                    InteractionEdge(
                        var_a=str(e["var_a"]),
                        var_b=str(e["var_b"]),
                        interaction_type=str(e["interaction_type"]),
                        description=str(e["description"]),
                    )
                )
        else:
            warnings.append("interactions: expected top-level 'edges' list")

    return Registry(entries=entries, interactions=edges, warnings=warnings)


def interactions_for(var: str, reg: Registry) -> list[dict[str, str]]:
    """
    Get all interactions for a given variable.

    Args:
        var: Variable name
        reg: Registry to search

    Returns:
        List of interaction dicts with keys: with, type, description
    """
    out = []
    for e in reg.interactions:
        if e.var_a == var or e.var_b == var:
            other = e.var_b if e.var_a == var else e.var_a
            out.append({"with": other, "type": e.interaction_type, "description": e.description})
    return out



"""HTML renderer for nim-audit output."""

from __future__ import annotations

from typing import Any

from jinja2 import Environment, BaseLoader

from nim_audit.renderers.base import BaseRenderer, OutputFormat, RenderContext


# Default HTML template
DEFAULT_TEMPLATE = """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{{ title }}</title>
    <style>
        :root {
            --bg-color: #1a1a2e;
            --card-bg: #16213e;
            --text-color: #eee;
            --accent-color: #0f3460;
            --success-color: #00bf63;
            --warning-color: #ffc107;
            --error-color: #dc3545;
            --info-color: #17a2b8;
        }

        * {
            box-sizing: border-box;
            margin: 0;
            padding: 0;
        }

        body {
            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, Oxygen, Ubuntu, sans-serif;
            background-color: var(--bg-color);
            color: var(--text-color);
            line-height: 1.6;
            padding: 2rem;
        }

        .container {
            max-width: 1200px;
            margin: 0 auto;
        }

        h1 {
            color: var(--text-color);
            margin-bottom: 1rem;
            border-bottom: 2px solid var(--accent-color);
            padding-bottom: 0.5rem;
        }

        h2 {
            color: var(--text-color);
            margin: 1.5rem 0 1rem;
        }

        .card {
            background-color: var(--card-bg);
            border-radius: 8px;
            padding: 1.5rem;
            margin-bottom: 1rem;
            box-shadow: 0 4px 6px rgba(0, 0, 0, 0.3);
        }

        .status-pass {
            color: var(--success-color);
            font-weight: bold;
        }

        .status-fail {
            color: var(--error-color);
            font-weight: bold;
        }

        .badge {
            display: inline-block;
            padding: 0.25rem 0.5rem;
            border-radius: 4px;
            font-size: 0.875rem;
            font-weight: 500;
        }

        .badge-error {
            background-color: var(--error-color);
        }

        .badge-warning {
            background-color: var(--warning-color);
            color: #000;
        }

        .badge-info {
            background-color: var(--info-color);
        }

        .badge-success {
            background-color: var(--success-color);
        }

        table {
            width: 100%;
            border-collapse: collapse;
            margin: 1rem 0;
        }

        th, td {
            padding: 0.75rem;
            text-align: left;
            border-bottom: 1px solid var(--accent-color);
        }

        th {
            background-color: var(--accent-color);
            font-weight: 600;
        }

        tr:hover {
            background-color: rgba(255, 255, 255, 0.05);
        }

        code {
            background-color: var(--accent-color);
            padding: 0.2rem 0.4rem;
            border-radius: 4px;
            font-family: 'Monaco', 'Menlo', monospace;
            font-size: 0.875rem;
        }

        .summary-grid {
            display: grid;
            grid-template-columns: repeat(auto-fit, minmax(150px, 1fr));
            gap: 1rem;
            margin: 1rem 0;
        }

        .summary-item {
            text-align: center;
            padding: 1rem;
            background-color: var(--accent-color);
            border-radius: 8px;
        }

        .summary-value {
            font-size: 2rem;
            font-weight: bold;
        }

        .summary-label {
            font-size: 0.875rem;
            opacity: 0.8;
        }

        .timestamp {
            font-size: 0.875rem;
            opacity: 0.7;
        }

        .breaking-change {
            border-left: 4px solid var(--error-color);
            padding-left: 1rem;
            margin: 1rem 0;
        }

        footer {
            margin-top: 2rem;
            text-align: center;
            opacity: 0.6;
            font-size: 0.875rem;
        }
    </style>
</head>
<body>
    <div class="container">
        {{ content }}

        <footer>
            Generated by nim-audit
        </footer>
    </div>
</body>
</html>
"""


class HTMLRenderer(BaseRenderer):
    """Renderer for HTML output format.

    Converts audit results to styled HTML reports.

    Example:
        renderer = HTMLRenderer()
        html_str = renderer.render(diff_report, context)
    """

    def __init__(self, template: str | None = None) -> None:
        """Initialize the HTML renderer.

        Args:
            template: Custom Jinja2 template string
        """
        self._template = template or DEFAULT_TEMPLATE
        self._env = Environment(loader=BaseLoader(), autoescape=True)

    @property
    def format(self) -> OutputFormat:
        """The output format this renderer produces."""
        return OutputFormat.HTML

    def render(self, data: Any, context: RenderContext) -> str:
        """Render data to an HTML string.

        Args:
            data: The data to render
            context: Rendering context

        Returns:
            HTML string
        """
        # Generate content based on data type
        if hasattr(data, "__class__"):
            class_name = data.__class__.__name__

            if class_name == "DiffReport":
                title = "NIM Diff Report"
                content = self._render_diff_report(data)
            elif class_name == "ConfigReport":
                title = "NIM Configuration Report"
                content = self._render_config_report(data)
            elif class_name == "CompatReport":
                title = "GPU Compatibility Report"
                content = self._render_compat_report(data)
            elif class_name == "LintResult":
                title = "Lint Report"
                content = self._render_lint_result(data)
            else:
                title = "Report"
                content = self._render_generic(data)
        else:
            title = "Report"
            content = self._render_generic(data)

        # Load custom template if provided
        template_str = self._template
        if context.template:
            try:
                from pathlib import Path

                template_str = Path(context.template).read_text()
            except Exception:
                pass

        template = self._env.from_string(template_str)
        return template.render(title=title, content=content)

    def _render_diff_report(self, report: Any) -> str:
        """Render diff report content."""
        breaking_html = ""
        if report.breaking_changes:
            breaking_items = "".join(
                f"""
                <div class="breaking-change">
                    <h3>{bc.title}</h3>
                    <p>{bc.description}</p>
                    <p><strong>Impact:</strong> {bc.impact}</p>
                    {f'<p><strong>Migration:</strong> {bc.migration}</p>' if bc.migration else ''}
                </div>
                """
                for bc in report.breaking_changes
            )
            breaking_html = f"<h2>Breaking Changes</h2>{breaking_items}"

        rows = "".join(
            f"""
            <tr>
                <td>{entry.category.value}</td>
                <td><span class="badge badge-{self._change_type_class(entry.change_type.value)}">{entry.change_type.value}</span></td>
                <td><code>{entry.path}</code></td>
                <td>{entry.old_value or '-'}</td>
                <td>{entry.new_value or '-'}</td>
            </tr>
            """
            for entry in report.entries
        )

        return f"""
        <h1>NIM Diff Report</h1>
        <p class="timestamp">Generated: {report.generated_at.isoformat()}</p>

        <div class="card">
            <h2>Images Compared</h2>
            <p><strong>Source:</strong> <code>{report.source_image.reference}</code></p>
            <p><strong>Target:</strong> <code>{report.target_image.reference}</code></p>
        </div>

        <div class="card">
            <h2>Summary</h2>
            <div class="summary-grid">
                <div class="summary-item">
                    <div class="summary-value">{report.total_changes}</div>
                    <div class="summary-label">Total Changes</div>
                </div>
                <div class="summary-item">
                    <div class="summary-value" style="color: var(--success-color)">{report.added_count}</div>
                    <div class="summary-label">Added</div>
                </div>
                <div class="summary-item">
                    <div class="summary-value" style="color: var(--error-color)">{report.removed_count}</div>
                    <div class="summary-label">Removed</div>
                </div>
                <div class="summary-item">
                    <div class="summary-value" style="color: var(--warning-color)">{report.modified_count}</div>
                    <div class="summary-label">Modified</div>
                </div>
                <div class="summary-item">
                    <div class="summary-value" style="color: var(--error-color)">{len(report.breaking_changes)}</div>
                    <div class="summary-label">Breaking</div>
                </div>
            </div>
        </div>

        {breaking_html}

        <div class="card">
            <h2>All Changes</h2>
            <table>
                <thead>
                    <tr>
                        <th>Category</th>
                        <th>Type</th>
                        <th>Path</th>
                        <th>Old Value</th>
                        <th>New Value</th>
                    </tr>
                </thead>
                <tbody>
                    {rows}
                </tbody>
            </table>
        </div>
        """

    def _render_config_report(self, report: Any) -> str:
        """Render config report content."""
        warnings_html = ""
        if report.warnings:
            items = "".join(f"<li>{w}</li>" for w in report.warnings)
            warnings_html = f'<div class="card"><h2>Warnings</h2><ul>{items}</ul></div>'

        rows = "".join(
            f"""
            <tr>
                <td><code>{entry.name}</code></td>
                <td>{entry.value or '-'}</td>
                <td>{entry.default_value or '-'}</td>
                <td><span class="badge badge-{self._impact_class(entry.impact.level.value if entry.impact else 'none')}">{entry.impact.level.value if entry.impact else '-'}</span></td>
                <td>{entry.description or '-'}</td>
            </tr>
            """
            for entry in report.entries
            if entry.is_set
        )

        return f"""
        <h1>NIM Configuration Report</h1>

        <div class="card">
            <p><strong>Image:</strong> <code>{report.image_reference}</code></p>
        </div>

        {warnings_html}

        <div class="card">
            <h2>Configuration</h2>
            <table>
                <thead>
                    <tr>
                        <th>Variable</th>
                        <th>Value</th>
                        <th>Default</th>
                        <th>Impact</th>
                        <th>Description</th>
                    </tr>
                </thead>
                <tbody>
                    {rows}
                </tbody>
            </table>
        </div>
        """

    def _render_compat_report(self, report: Any) -> str:
        """Render compatibility report content."""
        status_class = "status-pass" if report.compatible else "status-fail"
        status_text = "Compatible" if report.compatible else "Not Compatible"

        return f"""
        <h1>GPU Compatibility Report</h1>

        <div class="card">
            <p><strong>Image:</strong> <code>{report.image_reference}</code></p>
            <p><strong>Status:</strong> <span class="{status_class}">{status_text}</span></p>
        </div>

        <div class="card">
            <h2>Configuration</h2>
            <p><strong>GPU:</strong> {report.gpu.name if report.gpu else 'Not specified'}</p>
            <p><strong>Driver Version:</strong> {report.driver_version or 'Not specified'}</p>
        </div>

        <div class="card">
            <h2>Compatibility Checks</h2>
            <table>
                <tr>
                    <td>Compute Capability</td>
                    <td><span class="badge badge-{'success' if report.compute_compatible else 'error'}">{'Pass' if report.compute_compatible else 'Fail'}</span></td>
                </tr>
                <tr>
                    <td>GPU Memory</td>
                    <td><span class="badge badge-{'success' if report.memory_compatible else 'error'}">{'Pass' if report.memory_compatible else 'Fail'}</span></td>
                </tr>
                <tr>
                    <td>Driver Version</td>
                    <td><span class="badge badge-{'success' if report.driver_compatible else 'error'}">{'Pass' if report.driver_compatible else 'Fail'}</span></td>
                </tr>
            </table>
        </div>
        """

    def _render_lint_result(self, result: Any) -> str:
        """Render lint result content."""
        status_class = "status-pass" if result.passed else "status-fail"
        status_text = "Passed" if result.passed else "Failed"

        rows = "".join(
            f"""
            <tr>
                <td><span class="badge badge-{v.severity.value}">{v.severity.value}</span></td>
                <td>{v.rule.name}</td>
                <td>{v.message}</td>
                <td>{v.rule.remediation or '-'}</td>
            </tr>
            """
            for v in result.violations
        )

        return f"""
        <h1>Lint Report</h1>

        <div class="card">
            <p><strong>Image:</strong> <code>{result.image_reference}</code></p>
            <p><strong>Policy:</strong> {result.policy.name} v{result.policy.version}</p>
            <p><strong>Status:</strong> <span class="{status_class}">{status_text}</span></p>
        </div>

        <div class="card">
            <h2>Summary</h2>
            <div class="summary-grid">
                <div class="summary-item">
                    <div class="summary-value">{len(result.policy.enabled_rules)}</div>
                    <div class="summary-label">Total Rules</div>
                </div>
                <div class="summary-item">
                    <div class="summary-value" style="color: var(--error-color)">{result.error_count}</div>
                    <div class="summary-label">Errors</div>
                </div>
                <div class="summary-item">
                    <div class="summary-value" style="color: var(--warning-color)">{result.warning_count}</div>
                    <div class="summary-label">Warnings</div>
                </div>
            </div>
        </div>

        {'<div class="card"><h2>Violations</h2><table><thead><tr><th>Severity</th><th>Rule</th><th>Message</th><th>Remediation</th></tr></thead><tbody>' + rows + '</tbody></table></div>' if result.violations else '<div class="card"><p>No violations found!</p></div>'}
        """

    def _render_generic(self, data: Any) -> str:
        """Render generic data."""
        import json

        from pydantic import BaseModel

        if isinstance(data, BaseModel):
            dict_data = data.model_dump(mode="json")
        elif isinstance(data, dict):
            dict_data = data
        else:
            dict_data = {"data": str(data)}

        json_str = json.dumps(dict_data, indent=2, default=str)

        return f"""
        <h1>Report</h1>
        <div class="card">
            <pre><code>{json_str}</code></pre>
        </div>
        """

    @staticmethod
    def _change_type_class(change_type: str) -> str:
        """Get badge class for change type."""
        return {"added": "success", "removed": "error", "modified": "warning"}.get(
            change_type, "info"
        )

    @staticmethod
    def _impact_class(impact: str) -> str:
        """Get badge class for impact level."""
        return {
            "critical": "error",
            "high": "error",
            "medium": "warning",
            "low": "info",
        }.get(impact, "info")



"""Caching utilities for nim-audit."""

from __future__ import annotations

import hashlib
import json
import time
from pathlib import Path
from typing import Any, Callable, TypeVar

T = TypeVar("T")


class Cache:
    """Simple file-based cache for nim-audit.

    Caches expensive operations like image metadata fetching
    to speed up repeated audits.

    Example:
        cache = Cache()

        # Cache a value
        cache.set("key", {"data": "value"}, ttl=3600)

        # Get cached value
        value = cache.get("key")

        # Use decorator
        @cache.cached(ttl=3600)
        def expensive_operation(arg):
            return compute(arg)
    """

    def __init__(
        self,
        cache_dir: Path | str | None = None,
        default_ttl: int = 3600,
        enabled: bool = True,
    ) -> None:
        """Initialize the cache.

        Args:
            cache_dir: Directory for cache files. Defaults to ~/.cache/nim-audit
            default_ttl: Default time-to-live in seconds
            enabled: Whether caching is enabled
        """
        if cache_dir is None:
            cache_dir = Path.home() / ".cache" / "nim-audit"
        self._cache_dir = Path(cache_dir)
        self._default_ttl = default_ttl
        self._enabled = enabled
        self._memory_cache: dict[str, tuple[Any, float]] = {}

        # Create cache directory if it doesn't exist
        if self._enabled:
            self._cache_dir.mkdir(parents=True, exist_ok=True)

    @property
    def enabled(self) -> bool:
        """Whether caching is enabled."""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        """Enable or disable caching."""
        self._enabled = value

    def _key_to_path(self, key: str) -> Path:
        """Convert a cache key to a file path."""
        # Hash the key to create a safe filename
        key_hash = hashlib.sha256(key.encode()).hexdigest()[:32]
        return self._cache_dir / f"{key_hash}.json"

    def get(self, key: str, default: T | None = None) -> T | None:
        """Get a value from the cache.

        Args:
            key: Cache key
            default: Default value if not found or expired

        Returns:
            Cached value or default
        """
        if not self._enabled:
            return default

        # Check memory cache first
        if key in self._memory_cache:
            value, expires_at = self._memory_cache[key]
            if time.time() < expires_at:
                return value
            else:
                del self._memory_cache[key]

        # Check file cache
        cache_file = self._key_to_path(key)
        if cache_file.exists():
            try:
                data = json.loads(cache_file.read_text())
                if time.time() < data.get("expires_at", 0):
                    value = data["value"]
                    # Store in memory cache for faster access
                    self._memory_cache[key] = (value, data["expires_at"])
                    return value
                else:
                    # Expired - delete the file
                    cache_file.unlink(missing_ok=True)
            except (json.JSONDecodeError, KeyError, OSError):
                pass

        return default

    def set(self, key: str, value: Any, ttl: int | None = None) -> None:
        """Set a value in the cache.

        Args:
            key: Cache key
            value: Value to cache (must be JSON-serializable)
            ttl: Time-to-live in seconds. Uses default if not specified.
        """
        if not self._enabled:
            return

        if ttl is None:
            ttl = self._default_ttl

        expires_at = time.time() + ttl

        # Store in memory cache
        self._memory_cache[key] = (value, expires_at)

        # Store in file cache
        cache_file = self._key_to_path(key)
        try:
            data = {
                "key": key,
                "value": value,
                "expires_at": expires_at,
                "created_at": time.time(),
            }
            cache_file.write_text(json.dumps(data, default=str))
        except (OSError, TypeError):
            pass

    def delete(self, key: str) -> None:
        """Delete a value from the cache.

        Args:
            key: Cache key to delete
        """
        # Remove from memory cache
        self._memory_cache.pop(key, None)

        # Remove from file cache
        cache_file = self._key_to_path(key)
        cache_file.unlink(missing_ok=True)

    def clear(self) -> None:
        """Clear all cached values."""
        self._memory_cache.clear()

        if self._cache_dir.exists():
            for cache_file in self._cache_dir.glob("*.json"):
                cache_file.unlink(missing_ok=True)

    def cached(
        self,
        ttl: int | None = None,
        key_func: Callable[..., str] | None = None,
    ) -> Callable[[Callable[..., T]], Callable[..., T]]:
        """Decorator to cache function results.

        Args:
            ttl: Time-to-live in seconds
            key_func: Function to generate cache key from arguments

        Returns:
            Decorated function
        """

        def decorator(func: Callable[..., T]) -> Callable[..., T]:
            def wrapper(*args: Any, **kwargs: Any) -> T:
                if not self._enabled:
                    return func(*args, **kwargs)

                # Generate cache key
                if key_func:
                    cache_key = key_func(*args, **kwargs)
                else:
                    # Default key generation
                    key_parts = [func.__module__, func.__name__]
                    key_parts.extend(str(arg) for arg in args)
                    key_parts.extend(f"{k}={v}" for k, v in sorted(kwargs.items()))
                    cache_key = ":".join(key_parts)

                # Check cache
                result = self.get(cache_key)
                if result is not None:
                    return result

                # Call function and cache result
                result = func(*args, **kwargs)
                self.set(cache_key, result, ttl)
                return result

            return wrapper

        return decorator

    def get_stats(self) -> dict[str, Any]:
        """Get cache statistics.

        Returns:
            Dictionary with cache stats
        """
        file_count = 0
        total_size = 0

        if self._cache_dir.exists():
            for cache_file in self._cache_dir.glob("*.json"):
                file_count += 1
                total_size += cache_file.stat().st_size

        return {
            "enabled": self._enabled,
            "cache_dir": str(self._cache_dir),
            "memory_entries": len(self._memory_cache),
            "file_entries": file_count,
            "total_size_bytes": total_size,
        }


# Global cache instance
_default_cache: Cache | None = None


def get_cache() -> Cache:
    """Get the default cache instance.

    Returns:
        Default Cache instance
    """
    global _default_cache
    if _default_cache is None:
        _default_cache = Cache()
    return _default_cache



"""Image-related data models."""

from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field


class ImageDigest(BaseModel):
    """Container image digest information."""

    model_config = {"frozen": True}

    algorithm: str = Field(default="sha256", description="Hash algorithm")
    hash: str = Field(description="The digest hash value")

    def __str__(self) -> str:
        return f"{self.algorithm}:{self.hash}"

    @classmethod
    def from_string(cls, digest: str) -> "ImageDigest":
        """Parse a digest string like 'sha256:abc123...'."""
        if ":" in digest:
            algorithm, hash_value = digest.split(":", 1)
            return cls(algorithm=algorithm, hash=hash_value)
        return cls(hash=digest)


class LayerInfo(BaseModel):
    """Information about a container image layer."""

    model_config = {"frozen": True}

    digest: ImageDigest = Field(description="Layer digest")
    size: int = Field(description="Layer size in bytes")
    media_type: str = Field(description="Layer media type")
    created_by: str | None = Field(default=None, description="Command that created this layer")


class ImageManifest(BaseModel):
    """Container image manifest information."""

    model_config = {"frozen": True}

    schema_version: int = Field(description="Manifest schema version")
    media_type: str = Field(description="Manifest media type")
    digest: ImageDigest = Field(description="Manifest digest")
    config_digest: ImageDigest = Field(description="Config blob digest")
    layers: list[LayerInfo] = Field(default_factory=list, description="Image layers")
    annotations: dict[str, str] = Field(
        default_factory=dict,
        description="OCI annotations",
    )


class ImageMetadata(BaseModel):
    """Comprehensive metadata about a NIM container image."""

    model_config = {"frozen": True}

    # Identity
    reference: str = Field(description="Full image reference (registry/repo:tag)")
    repository: str = Field(description="Repository name")
    tag: str | None = Field(default=None, description="Image tag")
    digest: ImageDigest | None = Field(default=None, description="Image digest")

    # Manifest
    manifest: ImageManifest | None = Field(default=None, description="Image manifest")

    # Labels
    labels: dict[str, str] = Field(default_factory=dict, description="Container labels")

    # Build info
    created: datetime | None = Field(default=None, description="Image creation timestamp")
    architecture: str | None = Field(default=None, description="Target architecture")
    os: str | None = Field(default=None, description="Target operating system")

    # NIM-specific metadata
    nim_version: str | None = Field(default=None, description="NIM version")
    model_name: str | None = Field(default=None, description="Model name")
    model_version: str | None = Field(default=None, description="Model version")
    quantization: str | None = Field(default=None, description="Quantization type (fp16, int8, etc)")

    # Environment
    env: dict[str, str] = Field(default_factory=dict, description="Default environment variables")
    exposed_ports: list[int] = Field(default_factory=list, description="Exposed ports")

    # Entrypoint
    entrypoint: list[str] = Field(default_factory=list, description="Container entrypoint")
    cmd: list[str] = Field(default_factory=list, description="Container command")

    # Raw config for extensibility
    raw_config: dict[str, Any] = Field(
        default_factory=dict,
        description="Raw container config for additional inspection",
    )

    @property
    def full_reference(self) -> str:
        """Get the full image reference with tag or digest."""
        if self.digest:
            return f"{self.reference}@{self.digest}"
        return self.reference

    @property
    def total_size(self) -> int:
        """Calculate total image size from layers."""
        if self.manifest:
            return sum(layer.size for layer in self.manifest.layers)
        return 0



"""Unit tests for the SafeExpressionEvaluator."""

import pytest

from nim_audit.utils.expression import SafeExpressionEvaluator, safe_eval


class TestSafeExpressionEvaluator:
    """Tests for SafeExpressionEvaluator."""

    @pytest.fixture
    def evaluator(self):
        """Create an evaluator instance."""
        return SafeExpressionEvaluator()

    @pytest.fixture
    def sample_context(self):
        """Create a sample context for evaluation."""
        return {
            "labels": {
                "com.nvidia.nim.version": "1.5.0",
                "com.nvidia.nim.model.name": "llama3",
            },
            "env": {
                "NIM_SERVER_PORT": "8000",
                "NIM_LOG_LEVEL": "INFO",
            },
            "exposed_ports": [8000, 8080],
            "name": "test-image",
            "count": 42,
        }

    def test_evaluate_literal_string(self, evaluator):
        """Test evaluating string literals."""
        assert evaluator.evaluate("'hello'", {}) == "hello"
        assert evaluator.evaluate('"world"', {}) == "world"

    def test_evaluate_literal_numbers(self, evaluator):
        """Test evaluating number literals."""
        assert evaluator.evaluate("42", {}) == 42
        assert evaluator.evaluate("3.14", {}) == 3.14
        assert evaluator.evaluate("-5", {}) == -5

    def test_evaluate_literal_booleans(self, evaluator):
        """Test evaluating boolean literals."""
        assert evaluator.evaluate("True", {}) is True
        assert evaluator.evaluate("False", {}) is False

    def test_evaluate_none(self, evaluator):
        """Test evaluating None."""
        assert evaluator.evaluate("None", {}) is None

    def test_evaluate_variable(self, evaluator, sample_context):
        """Test evaluating variables from context."""
        assert evaluator.evaluate("name", sample_context) == "test-image"
        assert evaluator.evaluate("count", sample_context) == 42

    def test_evaluate_dict_get(self, evaluator, sample_context):
        """Test dict.get() method."""
        assert evaluator.evaluate(
            "labels.get('com.nvidia.nim.version')", sample_context
        ) == "1.5.0"
        assert evaluator.evaluate(
            "labels.get('nonexistent', 'default')", sample_context
        ) == "default"

    def test_evaluate_dict_subscript(self, evaluator, sample_context):
        """Test dict subscript access."""
        assert evaluator.evaluate("env['NIM_SERVER_PORT']", sample_context) == "8000"

    def test_evaluate_comparisons(self, evaluator, sample_context):
        """Test comparison operators."""
        assert evaluator.evaluate("count == 42", sample_context) is True
        assert evaluator.evaluate("count != 0", sample_context) is True
        assert evaluator.evaluate("count > 40", sample_context) is True
        assert evaluator.evaluate("count < 50", sample_context) is True
        assert evaluator.evaluate("count >= 42", sample_context) is True
        assert evaluator.evaluate("count <= 42", sample_context) is True

    def test_evaluate_in_operator(self, evaluator, sample_context):
        """Test 'in' operator."""
        assert evaluator.evaluate("8000 in exposed_ports", sample_context) is True
        assert evaluator.evaluate("9000 in exposed_ports", sample_context) is False
        assert evaluator.evaluate("'NIM_SERVER_PORT' in env", sample_context) is True

    def test_evaluate_boolean_and(self, evaluator, sample_context):
        """Test 'and' operator."""
        assert evaluator.evaluate("count > 0 and count < 100", sample_context) is True
        assert evaluator.evaluate("count > 0 and count < 10", sample_context) is False

    def test_evaluate_boolean_or(self, evaluator, sample_context):
        """Test 'or' operator."""
        assert evaluator.evaluate("count > 100 or count < 50", sample_context) is True
        assert evaluator.evaluate("count > 100 or count < 0", sample_context) is False

    def test_evaluate_boolean_not(self, evaluator, sample_context):
        """Test 'not' operator."""
        assert evaluator.evaluate("not False", {}) is True
        assert evaluator.evaluate("not count > 100", sample_context) is True

    def test_evaluate_is_none(self, evaluator, sample_context):
        """Test 'is None' and 'is not None'."""
        sample_context["value"] = None
        assert evaluator.evaluate("value is None", sample_context) is True
        assert evaluator.evaluate("name is not None", sample_context) is True

    def test_evaluate_string_methods(self, evaluator, sample_context):
        """Test allowed string methods."""
        assert evaluator.evaluate("name.startswith('test')", sample_context) is True
        assert evaluator.evaluate("name.endswith('image')", sample_context) is True
        assert evaluator.evaluate("name.upper()", sample_context) == "TEST-IMAGE"

    def test_evaluate_builtin_functions(self, evaluator, sample_context):
        """Test allowed builtin functions."""
        assert evaluator.evaluate("len(exposed_ports)", sample_context) == 2
        assert evaluator.evaluate("str(count)", sample_context) == "42"
        assert evaluator.evaluate("int('100')", {}) == 100
        assert evaluator.evaluate("max([1, 2, 3])", {}) == 3
        assert evaluator.evaluate("min([1, 2, 3])", {}) == 1
        assert evaluator.evaluate("sum([1, 2, 3])", {}) == 6
        assert evaluator.evaluate("abs(-5)", {}) == 5

    def test_evaluate_any_all(self, evaluator):
        """Test any() and all() functions."""
        assert evaluator.evaluate("all([True, True, True])", {}) is True
        assert evaluator.evaluate("all([True, False, True])", {}) is False
        assert evaluator.evaluate("any([False, False, True])", {}) is True
        assert evaluator.evaluate("any([False, False, False])", {}) is False

    def test_evaluate_arithmetic(self, evaluator, sample_context):
        """Test arithmetic operations."""
        assert evaluator.evaluate("count + 8", sample_context) == 50
        assert evaluator.evaluate("count - 2", sample_context) == 40
        assert evaluator.evaluate("count * 2", sample_context) == 84
        assert evaluator.evaluate("count / 2", sample_context) == 21.0
        assert evaluator.evaluate("count % 5", sample_context) == 2

    def test_evaluate_list_literal(self, evaluator):
        """Test list literals."""
        assert evaluator.evaluate("[1, 2, 3]", {}) == [1, 2, 3]
        assert evaluator.evaluate("['a', 'b']", {}) == ["a", "b"]

    def test_evaluate_dict_literal(self, evaluator):
        """Test dict literals."""
        assert evaluator.evaluate("{'a': 1, 'b': 2}", {}) == {"a": 1, "b": 2}

    def test_evaluate_ternary(self, evaluator, sample_context):
        """Test ternary if expression."""
        assert evaluator.evaluate("'high' if count > 10 else 'low'", sample_context) == "high"
        assert evaluator.evaluate("'low' if count > 100 else 'normal'", sample_context) == "normal"

    def test_evaluate_list_comprehension(self, evaluator):
        """Test simple list comprehension."""
        assert evaluator.evaluate("[x * 2 for x in [1, 2, 3]]", {}) == [2, 4, 6]
        assert evaluator.evaluate("[x for x in [1, 2, 3, 4] if x > 2]", {}) == [3, 4]

    def test_evaluate_chained_comparison(self, evaluator, sample_context):
        """Test chained comparisons."""
        assert evaluator.evaluate("10 < count < 100", sample_context) is True
        assert evaluator.evaluate("50 < count < 100", sample_context) is False

    def test_evaluate_complex_policy_condition(self, evaluator, sample_context):
        """Test complex conditions like those in policy rules."""
        # Rule: NIM version must exist
        assert evaluator.evaluate(
            "labels.get('com.nvidia.nim.version') is not None",
            sample_context,
        ) is True

        # Rule: Port 8000 must be exposed
        assert evaluator.evaluate(
            "8000 in exposed_ports or 80 in exposed_ports",
            sample_context,
        ) is True

        # Rule: No sensitive env vars (using list comprehension instead of generator)
        assert evaluator.evaluate(
            "not any([k in ['PASSWORD', 'SECRET'] for k in list(env.keys())])",
            sample_context,
        ) is True

    def test_evaluate_unknown_variable_raises(self, evaluator):
        """Test that unknown variables raise ValueError."""
        with pytest.raises(ValueError, match="Unknown variable"):
            evaluator.evaluate("unknown_var", {})

    def test_evaluate_invalid_syntax_raises(self, evaluator):
        """Test that invalid syntax raises ValueError."""
        with pytest.raises(ValueError, match="Invalid expression syntax"):
            evaluator.evaluate("1 +", {})

    def test_evaluate_deeply_nested_raises(self):
        """Test that deeply nested expressions raise ValueError."""
        evaluator = SafeExpressionEvaluator(max_depth=5)
        # Create a deeply nested expression with actual operations
        # Each operation adds depth, so we need nested operations
        deep = "[[[[[[[[1]]]]]]]]"  # Nested lists increase depth
        with pytest.raises(ValueError, match="too deeply nested"):
            evaluator.evaluate(deep, {})

    def test_safe_eval_function(self, sample_context):
        """Test the convenience safe_eval function."""
        assert safe_eval("count > 0", sample_context) is True
        assert safe_eval("labels.get('com.nvidia.nim.version')", sample_context) == "1.5.0"


class TestSafeEvalPolicyRules:
    """Test safe_eval with real policy rule conditions."""

    @pytest.fixture
    def image_context(self):
        """Create an image-like context."""
        return {
            "labels": {
                "com.nvidia.nim.version": "1.5.0",
                "com.nvidia.nim.model.name": "llama3-8b",
            },
            "env": {
                "NIM_SERVER_PORT": "8000",
                "NIM_MAX_BATCH_SIZE": "16",
            },
            "exposed_ports": [8000],
            "config": {"User": "nim"},
        }

    def test_require_version_label(self, image_context):
        """Test rule: require-version-label."""
        condition = "labels.get('com.nvidia.nim.version') is not None"
        assert safe_eval(condition, image_context) is True

        # Test failure case
        del image_context["labels"]["com.nvidia.nim.version"]
        assert safe_eval(condition, image_context) is False

    def test_no_root_user(self, image_context):
        """Test rule: no-root-user."""
        condition = "config.get('User', 'root') != 'root' and config.get('User', '') != ''"
        assert safe_eval(condition, image_context) is True

        # Test failure case
        image_context["config"]["User"] = "root"
        assert safe_eval(condition, image_context) is False

    def test_check_exposed_ports(self, image_context):
        """Test rule: check-exposed-ports."""
        condition = "8000 in exposed_ports or 80 in exposed_ports"
        assert safe_eval(condition, image_context) is True

        # Test failure case
        image_context["exposed_ports"] = [9000]
        assert safe_eval(condition, image_context) is False

    def test_batch_size_check(self, image_context):
        """Test rule: batch-size-check."""
        condition = "int(env.get('NIM_MAX_BATCH_SIZE', '1')) <= 64"
        assert safe_eval(condition, image_context) is True

        # Test failure case
        image_context["env"]["NIM_MAX_BATCH_SIZE"] = "128"
        assert safe_eval(condition, image_context) is False



"""Tokenizer extractor for NIM containers."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any

from nim_audit.extractors.base import ExtractorResult
from nim_audit.models.common import AuditError
from nim_audit.utils.hashing import hash_file


class TokenizerExtractor:
    """Extractor for tokenizer files from NIM containers.

    Extracts tokenizer configuration, vocabulary, and special tokens
    from NIM containers.

    Example:
        extractor = TokenizerExtractor()
        result = extractor.extract("nvcr.io/nim/llama3:1.5.0")
        print(result.data["vocab_size"])
    """

    # Tokenizer file patterns
    TOKENIZER_FILES = [
        "tokenizer.json",
        "tokenizer_config.json",
        "tokenizer.model",
        "vocab.json",
        "vocab.txt",
        "merges.txt",
        "special_tokens_map.json",
        "added_tokens.json",
        "spiece.model",
    ]

    # Common tokenizer directories
    TOKENIZER_DIRS = [
        "/opt/nim/models",
        "/models",
        "/workspace/models",
        "/home/nim/models",
    ]

    @property
    def name(self) -> str:
        """Unique name for this extractor."""
        return "tokenizer"

    @property
    def description(self) -> str:
        """Human-readable description."""
        return "Extracts tokenizer files, vocabulary, and special tokens configuration"

    def can_extract(self, image_id: str) -> bool:
        """Check if this extractor can handle the given image.

        Args:
            image_id: Image identifier

        Returns:
            True for NIM images
        """
        return "nim" in image_id.lower() or "nvcr.io" in image_id.lower()

    def extract(
        self,
        image_id: str,
        container_fs: Path | None = None,
    ) -> ExtractorResult:
        """Extract tokenizer information from the image.

        Args:
            image_id: Image identifier
            container_fs: Optional path to extracted container filesystem

        Returns:
            ExtractorResult with tokenizer information
        """
        try:
            data: dict[str, Any] = {
                "tokenizer_files": [],
                "tokenizer_type": None,
                "vocab_size": None,
                "special_tokens": {},
                "tokenizer_config": None,
                "bos_token": None,
                "eos_token": None,
                "pad_token": None,
                "unk_token": None,
            }

            if container_fs and container_fs.exists():
                data = self._extract_from_fs(container_fs)
            else:
                data = self._extract_from_image(image_id)

            return ExtractorResult.ok(self.name, data)

        except Exception as e:
            return ExtractorResult.fail(
                self.name,
                [
                    AuditError(
                        code="TOKENIZER_EXTRACTION_ERROR",
                        message=f"Failed to extract tokenizer info: {e}",
                        details={"image_id": image_id},
                    )
                ],
            )

    def _extract_from_fs(self, container_fs: Path) -> dict[str, Any]:
        """Extract tokenizer info from a filesystem path."""
        data: dict[str, Any] = {
            "tokenizer_files": [],
            "tokenizer_type": None,
            "vocab_size": None,
            "special_tokens": {},
            "tokenizer_config": None,
            "bos_token": None,
            "eos_token": None,
            "pad_token": None,
            "unk_token": None,
        }

        # Search for tokenizer files
        for tokenizer_dir in self.TOKENIZER_DIRS:
            dir_path = container_fs / tokenizer_dir.lstrip("/")
            if dir_path.exists():
                self._scan_directory(dir_path, data)

        return data

    def _extract_from_image(self, image_id: str) -> dict[str, Any]:
        """Extract tokenizer info by inspecting image."""
        data: dict[str, Any] = {
            "tokenizer_files": [],
            "tokenizer_type": None,
            "vocab_size": None,
            "special_tokens": {},
            "tokenizer_config": None,
            "bos_token": None,
            "eos_token": None,
            "pad_token": None,
            "unk_token": None,
        }

        try:
            import docker

            client = docker.from_env()
            container = client.containers.create(image_id, command="sleep 1")

            try:
                for tokenizer_dir in self.TOKENIZER_DIRS:
                    for tokenizer_file in self.TOKENIZER_FILES:
                        file_path = f"{tokenizer_dir}/{tokenizer_file}"
                        exit_code, content = container.exec_run(f"cat {file_path} 2>/dev/null")

                        if exit_code == 0 and content:
                            data["tokenizer_files"].append(file_path)

                            # Parse JSON files
                            if tokenizer_file.endswith(".json"):
                                try:
                                    file_data = json.loads(content.decode())
                                    self._process_tokenizer_file(tokenizer_file, file_data, data)
                                except json.JSONDecodeError:
                                    pass
            finally:
                container.remove(force=True)

        except ImportError:
            pass
        except Exception:
            pass

        return data

    def _scan_directory(self, dir_path: Path, data: dict[str, Any]) -> None:
        """Scan a directory for tokenizer files."""
        for path in dir_path.rglob("*"):
            if path.is_file() and path.name in self.TOKENIZER_FILES:
                file_info = {
                    "path": str(path),
                    "size": path.stat().st_size,
                    "hash": hash_file(path)[:16],
                }
                data["tokenizer_files"].append(file_info)

                # Parse JSON files
                if path.suffix == ".json":
                    try:
                        file_data = json.loads(path.read_text())
                        self._process_tokenizer_file(path.name, file_data, data)
                    except (json.JSONDecodeError, OSError):
                        pass

    def _process_tokenizer_file(
        self,
        filename: str,
        file_data: dict[str, Any],
        data: dict[str, Any],
    ) -> None:
        """Process a tokenizer file and extract information."""
        if filename == "tokenizer_config.json":
            data["tokenizer_config"] = file_data
            data["tokenizer_type"] = file_data.get("tokenizer_class")
            data["bos_token"] = self._extract_token(file_data.get("bos_token"))
            data["eos_token"] = self._extract_token(file_data.get("eos_token"))
            data["pad_token"] = self._extract_token(file_data.get("pad_token"))
            data["unk_token"] = self._extract_token(file_data.get("unk_token"))

        elif filename == "tokenizer.json":
            # HuggingFace tokenizers format
            if "model" in file_data:
                model = file_data["model"]
                if "vocab" in model:
                    data["vocab_size"] = len(model["vocab"])

            # Get added tokens
            if "added_tokens" in file_data:
                for token in file_data["added_tokens"]:
                    if isinstance(token, dict) and "content" in token:
                        data["special_tokens"][token["content"]] = token.get("id")

        elif filename == "special_tokens_map.json":
            data["special_tokens"].update(file_data)

        elif filename == "vocab.json":
            data["vocab_size"] = len(file_data)

    @staticmethod
    def _extract_token(token_data: Any) -> str | None:
        """Extract token string from various formats."""
        if token_data is None:
            return None
        if isinstance(token_data, str):
            return token_data
        if isinstance(token_data, dict):
            return token_data.get("content")
        return str(token_data)



"""CLI command for behavioral fingerprinting."""

from pathlib import Path
from typing import Optional

import typer
from rich.console import Console
from rich.table import Table
from rich.panel import Panel

app = typer.Typer(help="Generate and compare behavioral fingerprints.")
console = Console()


@app.callback(invoke_without_command=True)
def fingerprint_cmd(
    ctx: typer.Context,
    image: Optional[str] = typer.Option(
        None,
        "--image",
        "-i",
        help="NIM image reference to fingerprint",
    ),
    endpoint: Optional[str] = typer.Option(
        None,
        "--endpoint",
        "-e",
        help="NIM endpoint URL (if already running)",
    ),
    output: Optional[Path] = typer.Option(
        None,
        "--output",
        "-o",
        help="Output file for fingerprint",
    ),
    format: str = typer.Option(
        "terminal",
        "--format",
        "-f",
        help="Output format (terminal, json)",
    ),
) -> None:
    """
    Generate a behavioral fingerprint for a NIM image.

    Runs a standard suite of prompts against the model to capture
    its behavioral signature for comparison across versions.

    Example:
        nim-audit fingerprint --image nvcr.io/nim/llama3:1.6.0 --endpoint http://localhost:8000
    """
    if ctx.invoked_subcommand is not None:
        return

    if not image:
        console.print("[red]Error:[/red] --image is required")
        raise typer.Exit(1)

    if not endpoint:
        console.print("[red]Error:[/red] --endpoint is required (container auto-start not yet supported)")
        raise typer.Exit(1)

    from nim_audit.core.image import NIMImage
    from nim_audit.core.fingerprint import BehavioralFingerprinter

    with console.status("Loading image..."):
        try:
            nim_image = NIMImage.from_local(image)
        except ValueError:
            nim_image = NIMImage.from_registry(image)

    with console.status("Generating fingerprint..."):
        fingerprinter = BehavioralFingerprinter()
        result = fingerprinter.generate(nim_image, endpoint=endpoint)

    if not result.success:
        console.print("[red]Error:[/red] Failed to generate fingerprint")
        for error in result.errors:
            console.print(f"  {error}")
        raise typer.Exit(1)

    fingerprint = result.fingerprint
    if fingerprint is None:
        console.print("[red]Error:[/red] No fingerprint generated")
        raise typer.Exit(1)

    # Save fingerprint
    if output:
        fingerprinter.save_fingerprint(fingerprint, str(output))
        console.print(f"[green]Fingerprint saved to {output}[/green]")

    if format == "json":
        import json

        json_str = json.dumps(fingerprint.model_dump(mode="json"), indent=2, default=str)
        if not output:
            console.print(json_str)
    else:
        _print_terminal_report(fingerprint)


@app.command()
def compare(
    source: Path = typer.Argument(..., help="Source fingerprint file"),
    target: Path = typer.Argument(..., help="Target fingerprint file"),
    tolerance: float = typer.Option(
        0.05,
        "--tolerance",
        "-t",
        help="Similarity tolerance (0.0-1.0)",
    ),
    format: str = typer.Option(
        "terminal",
        "--format",
        "-f",
        help="Output format (terminal, json)",
    ),
) -> None:
    """
    Compare two behavioral fingerprints.

    Analyzes the similarity between two fingerprints and reports
    any behavioral differences.

    Example:
        nim-audit fingerprint compare v1.5.json v1.6.json --tolerance 0.05
    """
    from nim_audit.core.fingerprint import BehavioralFingerprinter

    if not source.exists():
        console.print(f"[red]Error:[/red] Source file not found: {source}")
        raise typer.Exit(1)

    if not target.exists():
        console.print(f"[red]Error:[/red] Target file not found: {target}")
        raise typer.Exit(1)

    with console.status("Loading fingerprints..."):
        fingerprinter = BehavioralFingerprinter()
        source_fp = fingerprinter.load_fingerprint(str(source))
        target_fp = fingerprinter.load_fingerprint(str(target))

    with console.status("Comparing fingerprints..."):
        comparison = fingerprinter.compare(source_fp, target_fp, tolerance=tolerance)

    if format == "json":
        import json

        output_data = {
            "source": source_fp.image_reference,
            "target": target_fp.image_reference,
            "similarity_score": comparison.similarity_score,
            "identical_responses": comparison.identical_responses,
            "different_responses": comparison.different_responses,
            "latency_change_percent": comparison.latency_change_percent,
            "is_similar": comparison.is_similar,
            "differences": comparison.response_diffs,
        }
        console.print(json.dumps(output_data, indent=2))
    else:
        _print_comparison_report(comparison, tolerance)

    # Exit with error if not similar
    if not comparison.is_similar:
        raise typer.Exit(1)


def _print_terminal_report(fingerprint) -> None:
    """Print fingerprint details."""
    console.print()
    console.print(
        Panel(
            f"[bold]Image:[/bold] {fingerprint.image_reference}\n"
            f"[bold]ID:[/bold] {fingerprint.fingerprint_id}\n"
            f"[bold]Generated:[/bold] {fingerprint.generated_at}\n"
            f"[bold]Prompts:[/bold] {len(fingerprint.responses)}",
            title="Behavioral Fingerprint",
        )
    )

    # Response summary
    console.print()
    table = Table(title="Responses")
    table.add_column("Prompt ID", style="bold")
    table.add_column("Latency (ms)")
    table.add_column("Tokens In")
    table.add_column("Tokens Out")
    table.add_column("Hash")

    for resp in fingerprint.responses:
        table.add_row(
            resp.prompt_id,
            f"{resp.latency_ms:.1f}",
            str(resp.tokens_in),
            str(resp.tokens_out),
            resp.response_hash or "-",
        )

    console.print(table)

    # Metrics
    console.print()
    console.print("[bold]Metrics[/bold]")
    console.print(f"  Average Latency: {fingerprint.avg_latency_ms:.1f}ms")
    console.print(f"  Total Tokens In: {fingerprint.total_tokens_in}")
    console.print(f"  Total Tokens Out: {fingerprint.total_tokens_out}")


def _print_comparison_report(comparison, tolerance: float) -> None:
    """Print comparison details."""
    console.print()

    # Status
    if comparison.is_similar:
        status = "[bold green]SIMILAR[/bold green]"
    else:
        status = "[bold red]DIFFERENT[/bold red]"

    console.print(
        Panel(
            f"[bold]Source:[/bold] {comparison.source.image_reference}\n"
            f"[bold]Target:[/bold] {comparison.target.image_reference}\n"
            f"[bold]Similarity:[/bold] {comparison.similarity_score:.1%}\n"
            f"[bold]Status:[/bold] {status}",
            title="Fingerprint Comparison",
        )
    )

    # Summary
    console.print()
    table = Table(title="Summary", show_header=False)
    table.add_column("Metric", style="bold")
    table.add_column("Value")
    table.add_row("Identical Responses", str(comparison.identical_responses))
    table.add_row("Different Responses", str(comparison.different_responses))
    table.add_row("Similarity Score", f"{comparison.similarity_score:.1%}")
    table.add_row("Tolerance", f"{tolerance:.1%}")

    latency_color = "green" if comparison.latency_change_percent <= 10 else "yellow"
    if comparison.latency_change_percent > 25:
        latency_color = "red"
    table.add_row(
        "Latency Change",
        f"[{latency_color}]{comparison.latency_change_percent:+.1f}%[/{latency_color}]",
    )
    console.print(table)

    # Differences
    if comparison.response_diffs:
        console.print()
        console.print("[bold]Response Differences[/bold]")
        for diff in comparison.response_diffs[:5]:  # Show first 5
            console.print(f"\n  [bold]{diff['prompt_id']}[/bold]")
            console.print(f"    Source: {diff['source'][:60]}...")
            console.print(f"    Target: {diff['target'][:60]}...")

        if len(comparison.response_diffs) > 5:
            console.print(f"\n  [dim]... and {len(comparison.response_diffs) - 5} more[/dim]")



"""Plugin system for nim-audit."""

from __future__ import annotations

import importlib
import importlib.util
import sys
from pathlib import Path
from typing import Any, Protocol, runtime_checkable

from nim_audit.utils.logging import get_logger

logger = get_logger("plugins")


@runtime_checkable
class Plugin(Protocol):
    """Protocol for nim-audit plugins.

    Plugins can extend nim-audit with custom extractors, renderers,
    registry clients, and CLI commands.

    To create a plugin:
    1. Create a Python module/package
    2. Implement the Plugin protocol
    3. Register the plugin in the config file or via CLI

    Example:
        class MyPlugin:
            name = "my-plugin"
            version = "1.0.0"

            def init(self, context: PluginContext) -> None:
                # Register custom extractors, renderers, etc.
                context.register_extractor(MyExtractor())

            def cleanup(self) -> None:
                # Cleanup resources
                pass
    """

    name: str
    version: str

    def init(self, context: "PluginContext") -> None:
        """Initialize the plugin.

        Args:
            context: Plugin context for registration
        """
        ...

    def cleanup(self) -> None:
        """Cleanup plugin resources."""
        ...


class PluginContext:
    """Context provided to plugins for registration.

    Provides methods to register custom components with nim-audit.
    """

    def __init__(self) -> None:
        """Initialize the plugin context."""
        from nim_audit.extractors import get_default_registry as get_extractor_registry

        self._extractor_registry = get_extractor_registry()
        self._renderers: dict[str, Any] = {}
        self._commands: list[Any] = []
        self._hooks: dict[str, list[Any]] = {}

    def register_extractor(self, extractor: Any) -> None:
        """Register a custom extractor.

        Args:
            extractor: Extractor instance implementing the Extractor protocol
        """
        self._extractor_registry.register(extractor)
        logger.debug(f"Registered extractor: {extractor.name}")

    def register_renderer(self, name: str, renderer: Any) -> None:
        """Register a custom renderer.

        Args:
            name: Renderer name
            renderer: Renderer instance
        """
        self._renderers[name] = renderer
        logger.debug(f"Registered renderer: {name}")

    def register_command(self, command: Any) -> None:
        """Register a custom CLI command.

        Args:
            command: Typer command function
        """
        self._commands.append(command)
        logger.debug("Registered custom command")

    def register_hook(self, event: str, callback: Any) -> None:
        """Register a hook callback.

        Args:
            event: Event name (e.g., "before_diff", "after_lint")
            callback: Callback function
        """
        if event not in self._hooks:
            self._hooks[event] = []
        self._hooks[event].append(callback)
        logger.debug(f"Registered hook for event: {event}")

    def get_hooks(self, event: str) -> list[Any]:
        """Get all hooks for an event.

        Args:
            event: Event name

        Returns:
            List of hook callbacks
        """
        return self._hooks.get(event, [])


class PluginManager:
    """Manager for loading and managing plugins.

    Example:
        manager = PluginManager()
        manager.load_plugin("my_plugin")
        manager.load_plugin_from_path("/path/to/plugin.py")
    """

    def __init__(self) -> None:
        """Initialize the plugin manager."""
        self._plugins: dict[str, Plugin] = {}
        self._context = PluginContext()

    @property
    def context(self) -> PluginContext:
        """Get the plugin context."""
        return self._context

    def load_plugin(self, module_name: str) -> None:
        """Load a plugin by module name.

        Args:
            module_name: Python module name (e.g., "my_plugin" or "package.plugin")

        Raises:
            ImportError: If module cannot be imported
            ValueError: If module doesn't implement Plugin protocol
        """
        try:
            module = importlib.import_module(module_name)
        except ImportError as e:
            raise ImportError(f"Failed to import plugin module '{module_name}': {e}")

        self._register_module_plugin(module, module_name)

    def load_plugin_from_path(self, path: Path | str) -> None:
        """Load a plugin from a file path.

        Args:
            path: Path to Python file

        Raises:
            FileNotFoundError: If file doesn't exist
            ValueError: If file doesn't implement Plugin protocol
        """
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"Plugin file not found: {path}")

        # Generate a unique module name
        module_name = f"nim_audit_plugin_{path.stem}"

        # Load the module
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None or spec.loader is None:
            raise ValueError(f"Cannot load plugin from: {path}")

        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        self._register_module_plugin(module, str(path))

    def _register_module_plugin(self, module: Any, source: str) -> None:
        """Register a plugin from a loaded module.

        Args:
            module: Loaded Python module
            source: Source identifier for error messages
        """
        # Look for a Plugin class or instance
        plugin: Plugin | None = None

        # Check for a 'plugin' attribute
        if hasattr(module, "plugin"):
            plugin = module.plugin
        # Check for a 'Plugin' class
        elif hasattr(module, "Plugin"):
            plugin_class = module.Plugin
            plugin = plugin_class()
        # Check for any class implementing Plugin protocol
        else:
            for name in dir(module):
                obj = getattr(module, name)
                if (
                    isinstance(obj, type)
                    and name != "Plugin"
                    and hasattr(obj, "name")
                    and hasattr(obj, "version")
                    and hasattr(obj, "init")
                ):
                    plugin = obj()
                    break

        if plugin is None:
            raise ValueError(
                f"No plugin found in '{source}'. "
                "Module must export a 'plugin' instance, 'Plugin' class, "
                "or a class implementing the Plugin protocol."
            )

        # Validate plugin
        if not hasattr(plugin, "name") or not hasattr(plugin, "version"):
            raise ValueError(f"Plugin from '{source}' missing required 'name' or 'version'")

        # Check for duplicate
        if plugin.name in self._plugins:
            raise ValueError(f"Plugin '{plugin.name}' is already loaded")

        # Initialize and register
        plugin.init(self._context)
        self._plugins[plugin.name] = plugin
        logger.info(f"Loaded plugin: {plugin.name} v{plugin.version}")

    def unload_plugin(self, name: str) -> None:
        """Unload a plugin by name.

        Args:
            name: Plugin name

        Raises:
            KeyError: If plugin is not loaded
        """
        if name not in self._plugins:
            raise KeyError(f"Plugin '{name}' is not loaded")

        plugin = self._plugins.pop(name)
        plugin.cleanup()
        logger.info(f"Unloaded plugin: {name}")

    def get_plugin(self, name: str) -> Plugin | None:
        """Get a loaded plugin by name.

        Args:
            name: Plugin name

        Returns:
            Plugin instance or None
        """
        return self._plugins.get(name)

    @property
    def loaded_plugins(self) -> list[str]:
        """Get names of all loaded plugins."""
        return list(self._plugins.keys())

    def cleanup_all(self) -> None:
        """Cleanup all loaded plugins."""
        for name in list(self._plugins.keys()):
            self.unload_plugin(name)


# Global plugin manager
_plugin_manager: PluginManager | None = None


def get_plugin_manager() -> PluginManager:
    """Get the global plugin manager.

    Returns:
        Global PluginManager instance
    """
    global _plugin_manager
    if _plugin_manager is None:
        _plugin_manager = PluginManager()
    return _plugin_manager



"""Extractor registry for managing and discovering extractors."""

from typing import Iterator

from nim_audit.extractors.base import Extractor


class ExtractorRegistry:
    """Registry for managing artifact extractors.

    The registry allows dynamic registration and discovery of extractors.
    It supports filtering extractors by capability.

    Example:
        registry = ExtractorRegistry()
        registry.register(MetadataExtractor())
        registry.register(ModelExtractor())

        # Get all extractors that can handle an image
        for extractor in registry.extractors_for("nvcr.io/nim/llama3:1.5.0"):
            result = extractor.extract(image_id)
    """

    def __init__(self) -> None:
        """Initialize an empty registry."""
        self._extractors: dict[str, Extractor] = {}

    def register(self, extractor: Extractor) -> None:
        """Register an extractor.

        Args:
            extractor: The extractor to register

        Raises:
            ValueError: If an extractor with the same name is already registered
        """
        if extractor.name in self._extractors:
            raise ValueError(f"Extractor '{extractor.name}' is already registered")
        self._extractors[extractor.name] = extractor

    def unregister(self, name: str) -> None:
        """Unregister an extractor by name.

        Args:
            name: Name of the extractor to unregister

        Raises:
            KeyError: If no extractor with that name is registered
        """
        if name not in self._extractors:
            raise KeyError(f"No extractor named '{name}' is registered")
        del self._extractors[name]

    def get(self, name: str) -> Extractor | None:
        """Get an extractor by name.

        Args:
            name: Name of the extractor

        Returns:
            The extractor or None if not found
        """
        return self._extractors.get(name)

    def __getitem__(self, name: str) -> Extractor:
        """Get an extractor by name.

        Args:
            name: Name of the extractor

        Returns:
            The extractor

        Raises:
            KeyError: If no extractor with that name is registered
        """
        if name not in self._extractors:
            raise KeyError(f"No extractor named '{name}' is registered")
        return self._extractors[name]

    def __contains__(self, name: str) -> bool:
        """Check if an extractor is registered.

        Args:
            name: Name of the extractor

        Returns:
            True if registered
        """
        return name in self._extractors

    def __iter__(self) -> Iterator[Extractor]:
        """Iterate over all registered extractors."""
        return iter(self._extractors.values())

    def __len__(self) -> int:
        """Get number of registered extractors."""
        return len(self._extractors)

    @property
    def names(self) -> list[str]:
        """Get names of all registered extractors."""
        return list(self._extractors.keys())

    def extractors_for(self, image_id: str) -> Iterator[Extractor]:
        """Get extractors that can handle the given image.

        Args:
            image_id: Image identifier

        Yields:
            Extractors that can extract from the image
        """
        for extractor in self._extractors.values():
            if extractor.can_extract(image_id):
                yield extractor

    def clear(self) -> None:
        """Remove all registered extractors."""
        self._extractors.clear()


# Global default registry
_default_registry: ExtractorRegistry | None = None


def get_default_registry() -> ExtractorRegistry:
    """Get the default global extractor registry.

    Returns:
        The default ExtractorRegistry instance
    """
    global _default_registry
    if _default_registry is None:
        _default_registry = ExtractorRegistry()
    return _default_registry



"""Safe expression evaluator for policy rules."""

from __future__ import annotations

import ast
import operator
from typing import Any, Callable


class SafeExpressionEvaluator:
    """Safe evaluator for policy rule expressions.

    This evaluator uses Python's AST to parse and evaluate expressions
    safely, without using eval(). It supports a limited subset of Python
    expressions suitable for policy rules.

    Supported operations:
    - Comparisons: ==, !=, <, <=, >, >=, in, not in, is, is not
    - Boolean: and, or, not
    - Attribute access: obj.attr
    - Subscript: obj[key], obj.get(key)
    - Method calls: str.startswith(), str.endswith(), etc.
    - Literals: strings, numbers, booleans, None, lists, dicts

    Example:
        evaluator = SafeExpressionEvaluator()
        context = {"labels": {"version": "1.0"}, "env": {"PORT": "8000"}}
        result = evaluator.evaluate("labels.get('version') == '1.0'", context)
    """

    # Allowed comparison operators
    COMPARE_OPS: dict[type, Callable[[Any, Any], bool]] = {
        ast.Eq: operator.eq,
        ast.NotEq: operator.ne,
        ast.Lt: operator.lt,
        ast.LtE: operator.le,
        ast.Gt: operator.gt,
        ast.GtE: operator.ge,
        ast.Is: operator.is_,
        ast.IsNot: operator.is_not,
        ast.In: lambda a, b: a in b,
        ast.NotIn: lambda a, b: a not in b,
    }

    # Allowed binary operators
    BINARY_OPS: dict[type, Callable[[Any, Any], Any]] = {
        ast.Add: operator.add,
        ast.Sub: operator.sub,
        ast.Mult: operator.mul,
        ast.Div: operator.truediv,
        ast.Mod: operator.mod,
    }

    # Allowed unary operators
    UNARY_OPS: dict[type, Callable[[Any], Any]] = {
        ast.Not: operator.not_,
        ast.USub: operator.neg,
        ast.UAdd: operator.pos,
    }

    # Allowed built-in functions
    ALLOWED_FUNCTIONS: dict[str, Callable[..., Any]] = {
        "len": len,
        "str": str,
        "int": int,
        "float": float,
        "bool": bool,
        "list": list,
        "dict": dict,
        "set": set,
        "min": min,
        "max": max,
        "sum": sum,
        "abs": abs,
        "all": all,
        "any": any,
        "sorted": sorted,
        "reversed": lambda x: list(reversed(x)),
    }

    # Allowed methods on strings
    ALLOWED_STRING_METHODS = {
        "startswith",
        "endswith",
        "lower",
        "upper",
        "strip",
        "lstrip",
        "rstrip",
        "split",
        "join",
        "replace",
        "find",
        "count",
        "isdigit",
        "isalpha",
        "isalnum",
    }

    # Allowed methods on dicts
    ALLOWED_DICT_METHODS = {
        "get",
        "keys",
        "values",
        "items",
    }

    # Allowed methods on lists
    ALLOWED_LIST_METHODS = {
        "append",
        "extend",
        "index",
        "count",
    }

    def __init__(self, max_depth: int = 10) -> None:
        """Initialize the evaluator.

        Args:
            max_depth: Maximum AST depth to prevent stack overflow
        """
        self._max_depth = max_depth

    def evaluate(self, expression: str, context: dict[str, Any]) -> Any:
        """Evaluate an expression with the given context.

        Args:
            expression: Python expression string
            context: Variables available in the expression

        Returns:
            Result of evaluating the expression

        Raises:
            ValueError: If expression is invalid or unsafe
        """
        try:
            tree = ast.parse(expression, mode="eval")
        except SyntaxError as e:
            raise ValueError(f"Invalid expression syntax: {e}")

        return self._eval_node(tree.body, context, depth=0)

    def _eval_node(self, node: ast.AST, context: dict[str, Any], depth: int) -> Any:
        """Evaluate an AST node.

        Args:
            node: AST node to evaluate
            context: Variable context
            depth: Current recursion depth

        Returns:
            Result of evaluation
        """
        if depth > self._max_depth:
            raise ValueError("Expression too deeply nested")

        # Constants (Python 3.8+)
        if isinstance(node, ast.Constant):
            return node.value

        # Names (variables)
        if isinstance(node, ast.Name):
            if node.id in self.ALLOWED_FUNCTIONS:
                return self.ALLOWED_FUNCTIONS[node.id]
            if node.id == "True":
                return True
            if node.id == "False":
                return False
            if node.id == "None":
                return None
            if node.id in context:
                return context[node.id]
            raise ValueError(f"Unknown variable: {node.id}")

        # Attribute access (obj.attr)
        if isinstance(node, ast.Attribute):
            value = self._eval_node(node.value, context, depth + 1)
            attr = node.attr

            # Check if it's an allowed method
            if isinstance(value, str) and attr in self.ALLOWED_STRING_METHODS:
                return getattr(value, attr)
            if isinstance(value, dict) and attr in self.ALLOWED_DICT_METHODS:
                return getattr(value, attr)
            if isinstance(value, list) and attr in self.ALLOWED_LIST_METHODS:
                return getattr(value, attr)

            # Allow attribute access on objects
            if hasattr(value, attr):
                return getattr(value, attr)

            raise ValueError(f"Cannot access attribute '{attr}'")

        # Subscript (obj[key])
        if isinstance(node, ast.Subscript):
            value = self._eval_node(node.value, context, depth + 1)
            key = self._eval_node(node.slice, context, depth + 1)
            return value[key]

        # Function/method calls
        if isinstance(node, ast.Call):
            func = self._eval_node(node.func, context, depth + 1)

            # Evaluate arguments
            args = [self._eval_node(arg, context, depth + 1) for arg in node.args]
            kwargs = {
                kw.arg: self._eval_node(kw.value, context, depth + 1)
                for kw in node.keywords
                if kw.arg is not None
            }

            # Check if it's a safe callable
            if callable(func):
                return func(*args, **kwargs)

            raise ValueError("Cannot call non-callable")

        # Comparisons
        if isinstance(node, ast.Compare):
            left = self._eval_node(node.left, context, depth + 1)

            for op, comparator in zip(node.ops, node.comparators):
                op_func = self.COMPARE_OPS.get(type(op))
                if op_func is None:
                    raise ValueError(f"Unsupported comparison operator: {type(op).__name__}")

                right = self._eval_node(comparator, context, depth + 1)
                if not op_func(left, right):
                    return False
                left = right

            return True

        # Boolean operations (and, or)
        if isinstance(node, ast.BoolOp):
            if isinstance(node.op, ast.And):
                for value in node.values:
                    if not self._eval_node(value, context, depth + 1):
                        return False
                return True
            elif isinstance(node.op, ast.Or):
                for value in node.values:
                    if self._eval_node(value, context, depth + 1):
                        return True
                return False

        # Unary operations (not, -, +)
        if isinstance(node, ast.UnaryOp):
            op_func = self.UNARY_OPS.get(type(node.op))
            if op_func is None:
                raise ValueError(f"Unsupported unary operator: {type(node.op).__name__}")
            operand = self._eval_node(node.operand, context, depth + 1)
            return op_func(operand)

        # Binary operations (+, -, *, /, %)
        if isinstance(node, ast.BinOp):
            op_func = self.BINARY_OPS.get(type(node.op))
            if op_func is None:
                raise ValueError(f"Unsupported binary operator: {type(node.op).__name__}")
            left = self._eval_node(node.left, context, depth + 1)
            right = self._eval_node(node.right, context, depth + 1)
            return op_func(left, right)

        # List literals
        if isinstance(node, ast.List):
            return [self._eval_node(elt, context, depth + 1) for elt in node.elts]

        # Dict literals
        if isinstance(node, ast.Dict):
            return {
                self._eval_node(k, context, depth + 1): self._eval_node(v, context, depth + 1)
                for k, v in zip(node.keys, node.values)
                if k is not None
            }

        # Set literals
        if isinstance(node, ast.Set):
            return {self._eval_node(elt, context, depth + 1) for elt in node.elts}

        # Tuple literals
        if isinstance(node, ast.Tuple):
            return tuple(self._eval_node(elt, context, depth + 1) for elt in node.elts)

        # If expressions (ternary)
        if isinstance(node, ast.IfExp):
            test = self._eval_node(node.test, context, depth + 1)
            if test:
                return self._eval_node(node.body, context, depth + 1)
            else:
                return self._eval_node(node.orelse, context, depth + 1)

        # List comprehensions (limited)
        if isinstance(node, ast.ListComp):
            return self._eval_comprehension(node, context, depth)

        raise ValueError(f"Unsupported expression type: {type(node).__name__}")

    def _eval_comprehension(
        self, node: ast.ListComp, context: dict[str, Any], depth: int
    ) -> list[Any]:
        """Evaluate a list comprehension.

        Only supports simple single-loop comprehensions.
        """
        if len(node.generators) != 1:
            raise ValueError("Only single-loop comprehensions are supported")

        gen = node.generators[0]
        if not isinstance(gen.target, ast.Name):
            raise ValueError("Only simple loop variables are supported")

        var_name = gen.target.id
        iterable = self._eval_node(gen.iter, context, depth + 1)

        result = []
        for item in iterable:
            local_context = {**context, var_name: item}

            # Check conditions
            if all(self._eval_node(cond, local_context, depth + 1) for cond in gen.ifs):
                result.append(self._eval_node(node.elt, local_context, depth + 1))

        return result


# Global evaluator instance
_evaluator = SafeExpressionEvaluator()


def safe_eval(expression: str, context: dict[str, Any]) -> Any:
    """Safely evaluate an expression.

    Args:
        expression: Python expression string
        context: Variables available in the expression

    Returns:
        Result of evaluating the expression
    """
    return _evaluator.evaluate(expression, context)



"""CLI command for GPU compatibility checking."""

from pathlib import Path
from typing import Optional

import typer
from rich.console import Console
from rich.table import Table
from rich.panel import Panel

console = Console()


def compat_cmd(
    image: str = typer.Option(..., "--image", "-i", help="NIM image reference"),
    gpu: Optional[str] = typer.Option(
        None,
        "--gpu",
        "-g",
        help="Target GPU (e.g., A100, H100, L4)",
    ),
    driver: Optional[str] = typer.Option(
        None,
        "--driver",
        "-d",
        help="NVIDIA driver version",
    ),
    cuda: Optional[str] = typer.Option(
        None,
        "--cuda",
        "-c",
        help="CUDA version",
    ),
    format: str = typer.Option(
        "terminal",
        "--format",
        "-f",
        help="Output format (terminal, json)",
    ),
    output: Optional[Path] = typer.Option(
        None,
        "--output",
        "-o",
        help="Output file path",
    ),
    detect: bool = typer.Option(
        False,
        "--detect",
        help="Auto-detect local GPU and driver",
    ),
) -> None:
    """
    Check GPU compatibility for a NIM image.

    Verifies that the target GPU meets the requirements
    for running the specified NIM container.

    Example:
        nim-audit compat --image nvcr.io/nim/llama3:1.6.0 --gpu A10 --driver 550.54
    """
    from nim_audit.core.image import NIMImage
    from nim_audit.core.compat import CompatChecker

    with console.status("Loading image..."):
        try:
            nim_image = NIMImage.from_local(image)
        except ValueError:
            nim_image = NIMImage.from_registry(image)

    # Auto-detect GPU if requested
    detected_gpu = None
    detected_driver = None
    if detect:
        with console.status("Detecting GPU..."):
            detected_gpu, detected_driver = _detect_gpu()
            if detected_gpu:
                console.print(f"[green]Detected GPU:[/green] {detected_gpu}")
            if detected_driver:
                console.print(f"[green]Detected Driver:[/green] {detected_driver}")

    # Use detected values if not explicitly provided
    gpu = gpu or detected_gpu
    driver = driver or detected_driver

    if not gpu:
        console.print("[yellow]Warning:[/yellow] No GPU specified. Use --gpu or --detect")

    with console.status("Checking compatibility..."):
        checker = CompatChecker()
        result = checker.check(nim_image, gpu=gpu, driver_version=driver, cuda_version=cuda)

    if not result.success:
        console.print("[red]Error:[/red] Failed to check compatibility")
        for error in result.errors:
            console.print(f"  {error}")
        raise typer.Exit(1)

    report = result.report
    if report is None:
        console.print("[red]Error:[/red] No report generated")
        raise typer.Exit(1)

    if format == "json":
        import json

        output_data = report.model_dump(mode="json")
        json_str = json.dumps(output_data, indent=2, default=str)

        if output:
            output.write_text(json_str)
            console.print(f"Report written to {output}")
        else:
            console.print(json_str)
    else:
        _print_terminal_report(report)

    # Exit with error code if not compatible
    if not report.compatible:
        raise typer.Exit(1)


def _print_terminal_report(report) -> None:
    """Print a rich terminal report."""
    console.print()

    # Compatibility status
    if report.compatible:
        status = "[bold green]COMPATIBLE[/bold green]"
    else:
        status = "[bold red]NOT COMPATIBLE[/bold red]"

    console.print(
        Panel(
            f"[bold]Image:[/bold] {report.image_reference}\n"
            f"[bold]GPU:[/bold] {report.gpu.name if report.gpu else 'Not specified'}\n"
            f"[bold]Driver:[/bold] {report.driver_version or 'Not specified'}\n"
            f"[bold]Status:[/bold] {status}",
            title="GPU Compatibility Check",
        )
    )

    # Requirements
    reqs = report.requirements
    console.print()
    table = Table(title="Requirements")
    table.add_column("Requirement", style="bold")
    table.add_column("Required")
    table.add_column("Actual")
    table.add_column("Status")

    def status_icon(ok: bool) -> str:
        return "[green]OK[/green]" if ok else "[red]FAIL[/red]"

    if reqs.min_compute_capability:
        actual = report.gpu.compute_capability if report.gpu else "Unknown"
        table.add_row(
            "Compute Capability",
            f">= {reqs.min_compute_capability}",
            actual or "Unknown",
            status_icon(report.compute_compatible),
        )

    if reqs.min_memory_gb:
        actual = f"{report.gpu.memory_gb}GB" if report.gpu and report.gpu.memory_gb else "Unknown"
        table.add_row(
            "GPU Memory",
            f">= {reqs.min_memory_gb}GB",
            actual,
            status_icon(report.memory_compatible),
        )

    if reqs.min_driver_version:
        table.add_row(
            "Driver Version",
            f">= {reqs.min_driver_version}",
            report.driver_version or "Unknown",
            status_icon(report.driver_compatible),
        )

    if reqs.supported_gpus:
        gpu_name = report.gpu.name if report.gpu else "Unknown"
        table.add_row(
            "Supported GPUs",
            ", ".join(reqs.supported_gpus[:3]) + ("..." if len(reqs.supported_gpus) > 3 else ""),
            gpu_name,
            status_icon(report.gpu_supported),
        )

    console.print(table)

    # Warnings
    if report.warnings:
        console.print()
        console.print("[bold yellow]Warnings[/bold yellow]")
        for warning in report.warnings:
            console.print(f"  [yellow]![/yellow] {warning}")

    # Recommendations
    if report.recommendations:
        console.print()
        console.print("[bold]Recommendations[/bold]")
        for rec in report.recommendations:
            console.print(f"  - {rec}")


def _detect_gpu() -> tuple[Optional[str], Optional[str]]:
    """Detect local GPU and driver version.

    Returns:
        Tuple of (gpu_name, driver_version)
    """
    try:
        import subprocess

        # Try nvidia-smi
        result = subprocess.run(
            ["nvidia-smi", "--query-gpu=name,driver_version", "--format=csv,noheader"],
            capture_output=True,
            text=True,
        )
        if result.returncode == 0:
            line = result.stdout.strip().split("\n")[0]
            parts = line.split(",")
            if len(parts) >= 2:
                gpu_name = parts[0].strip()
                # Extract model name (e.g., "NVIDIA A100" -> "A100")
                for prefix in ["NVIDIA ", "GeForce ", "Tesla "]:
                    if gpu_name.startswith(prefix):
                        gpu_name = gpu_name[len(prefix) :]
                driver_version = parts[1].strip()
                return gpu_name, driver_version
    except FileNotFoundError:
        pass
    except Exception:
        pass

    return None, None



"""DiffEngine for comparing NIM container images."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from nim_audit.models.common import AuditError
from nim_audit.models.diff import (
    BreakingChange,
    ChangeCategory,
    ChangeType,
    DiffEntry,
    DiffReport,
    DiffResult,
    Severity,
)

if TYPE_CHECKING:
    from nim_audit.core.image import NIMImage
    from nim_audit.extractors.base import Extractor
    from nim_audit.renderers.base import Renderer


class DiffEngine:
    """Engine for comparing two NIM container images.

    The DiffEngine identifies differences between container images,
    including metadata, model artifacts, tokenizers, and API schemas.
    It also detects breaking changes that may affect compatibility.

    Example:
        engine = DiffEngine()
        result = engine.diff(image1, image2)

        if result.success:
            for entry in result.report.entries:
                print(f"{entry.path}: {entry.change_type}")
            for bc in result.report.breaking_changes:
                print(f"BREAKING: {bc.title}")
    """

    def __init__(
        self,
        extractors: list[Extractor] | None = None,
        renderer: Renderer | None = None,
    ) -> None:
        """Initialize the diff engine.

        Args:
            extractors: Optional list of artifact extractors to use
            renderer: Optional renderer for output
        """
        self._extractors = extractors or []
        self._renderer = renderer

    def diff(self, source: "NIMImage", target: "NIMImage") -> DiffResult:
        """Compare two NIM images and generate a diff report.

        Args:
            source: The source (old) image
            target: The target (new) image

        Returns:
            DiffResult containing the comparison report or errors
        """
        try:
            entries: list[DiffEntry] = []
            breaking_changes: list[BreakingChange] = []

            # Compare metadata
            metadata_entries, metadata_breaking = self._diff_metadata(source, target)
            entries.extend(metadata_entries)
            breaking_changes.extend(metadata_breaking)

            # Compare environment variables
            env_entries, env_breaking = self._diff_environment(source, target)
            entries.extend(env_entries)
            breaking_changes.extend(env_breaking)

            # Compare labels
            label_entries = self._diff_labels(source, target)
            entries.extend(label_entries)

            # Compare layers (structural)
            layer_entries, layer_breaking = self._diff_layers(source, target)
            entries.extend(layer_entries)
            breaking_changes.extend(layer_breaking)

            # Run extractors and compare extracted data
            for extractor in self._extractors:
                if extractor.can_extract(source.reference) and extractor.can_extract(
                    target.reference
                ):
                    source_result = extractor.extract(source.reference)
                    target_result = extractor.extract(target.reference)

                    if source_result.success and target_result.success:
                        extractor_entries = self._diff_extracted_data(
                            source_result.data,
                            target_result.data,
                            extractor.name,
                        )
                        entries.extend(extractor_entries)

            # Calculate statistics
            added = sum(1 for e in entries if e.change_type == ChangeType.ADDED)
            removed = sum(1 for e in entries if e.change_type == ChangeType.REMOVED)
            modified = sum(1 for e in entries if e.change_type == ChangeType.MODIFIED)

            report = DiffReport(
                source_image=source.metadata,
                target_image=target.metadata,
                generated_at=datetime.utcnow(),
                entries=entries,
                breaking_changes=breaking_changes,
                total_changes=len(entries),
                added_count=added,
                removed_count=removed,
                modified_count=modified,
            )

            return DiffResult.ok(report)

        except Exception as e:
            return DiffResult.fail(
                [
                    AuditError(
                        code="DIFF_ERROR",
                        message=f"Failed to diff images: {e}",
                        details={"source": source.reference, "target": target.reference},
                    )
                ]
            )

    def _diff_metadata(
        self, source: "NIMImage", target: "NIMImage"
    ) -> tuple[list[DiffEntry], list[BreakingChange]]:
        """Compare image metadata."""
        entries: list[DiffEntry] = []
        breaking: list[BreakingChange] = []

        sm = source.metadata
        tm = target.metadata

        # Compare NIM version
        if sm.nim_version != tm.nim_version:
            entry = DiffEntry(
                category=ChangeCategory.METADATA,
                change_type=ChangeType.MODIFIED,
                path="nim_version",
                old_value=sm.nim_version,
                new_value=tm.nim_version,
                severity=Severity.WARNING,
                description="NIM version changed",
            )
            entries.append(entry)

        # Compare model name
        if sm.model_name != tm.model_name:
            entry = DiffEntry(
                category=ChangeCategory.MODEL,
                change_type=ChangeType.MODIFIED,
                path="model_name",
                old_value=sm.model_name,
                new_value=tm.model_name,
                severity=Severity.BREAKING,
                description="Model name changed",
            )
            entries.append(entry)
            breaking.append(
                BreakingChange(
                    category=ChangeCategory.MODEL,
                    title="Model name changed",
                    description=f"Model changed from {sm.model_name} to {tm.model_name}",
                    impact="Different model may produce different outputs",
                    migration="Verify model outputs match expectations",
                    related_entries=["model_name"],
                )
            )

        # Compare model version
        if sm.model_version != tm.model_version:
            entry = DiffEntry(
                category=ChangeCategory.MODEL,
                change_type=ChangeType.MODIFIED,
                path="model_version",
                old_value=sm.model_version,
                new_value=tm.model_version,
                severity=Severity.WARNING,
                description="Model version changed",
            )
            entries.append(entry)

        # Compare quantization
        if sm.quantization != tm.quantization:
            entry = DiffEntry(
                category=ChangeCategory.MODEL,
                change_type=ChangeType.MODIFIED,
                path="quantization",
                old_value=sm.quantization,
                new_value=tm.quantization,
                severity=Severity.WARNING,
                description="Quantization method changed",
            )
            entries.append(entry)

        # Compare architecture
        if sm.architecture != tm.architecture:
            entry = DiffEntry(
                category=ChangeCategory.METADATA,
                change_type=ChangeType.MODIFIED,
                path="architecture",
                old_value=sm.architecture,
                new_value=tm.architecture,
                severity=Severity.BREAKING,
                description="Target architecture changed",
            )
            entries.append(entry)
            breaking.append(
                BreakingChange(
                    category=ChangeCategory.METADATA,
                    title="Architecture changed",
                    description=f"Architecture changed from {sm.architecture} to {tm.architecture}",
                    impact="Container may not run on previous hardware",
                    migration="Ensure target system supports new architecture",
                    related_entries=["architecture"],
                )
            )

        # Compare exposed ports
        source_ports = set(sm.exposed_ports)
        target_ports = set(tm.exposed_ports)

        for port in source_ports - target_ports:
            entries.append(
                DiffEntry(
                    category=ChangeCategory.CONFIG,
                    change_type=ChangeType.REMOVED,
                    path=f"exposed_ports/{port}",
                    old_value=str(port),
                    new_value=None,
                    severity=Severity.WARNING,
                    description=f"Port {port} no longer exposed",
                )
            )

        for port in target_ports - source_ports:
            entries.append(
                DiffEntry(
                    category=ChangeCategory.CONFIG,
                    change_type=ChangeType.ADDED,
                    path=f"exposed_ports/{port}",
                    old_value=None,
                    new_value=str(port),
                    severity=Severity.INFO,
                    description=f"Port {port} now exposed",
                )
            )

        return entries, breaking

    def _diff_environment(
        self, source: "NIMImage", target: "NIMImage"
    ) -> tuple[list[DiffEntry], list[BreakingChange]]:
        """Compare environment variables."""
        entries: list[DiffEntry] = []
        breaking: list[BreakingChange] = []

        source_env = source.metadata.env
        target_env = target.metadata.env

        all_keys = set(source_env.keys()) | set(target_env.keys())

        for key in sorted(all_keys):
            source_val = source_env.get(key)
            target_val = target_env.get(key)

            if source_val is None and target_val is not None:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.ENVIRONMENT,
                        change_type=ChangeType.ADDED,
                        path=f"env/{key}",
                        old_value=None,
                        new_value=target_val,
                        severity=Severity.INFO,
                        description=f"Environment variable {key} added",
                    )
                )
            elif source_val is not None and target_val is None:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.ENVIRONMENT,
                        change_type=ChangeType.REMOVED,
                        path=f"env/{key}",
                        old_value=source_val,
                        new_value=None,
                        severity=Severity.WARNING,
                        description=f"Environment variable {key} removed",
                    )
                )
            elif source_val != target_val:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.ENVIRONMENT,
                        change_type=ChangeType.MODIFIED,
                        path=f"env/{key}",
                        old_value=source_val,
                        new_value=target_val,
                        severity=Severity.INFO,
                        description=f"Environment variable {key} changed",
                    )
                )

        return entries, breaking

    def _diff_labels(self, source: "NIMImage", target: "NIMImage") -> list[DiffEntry]:
        """Compare container labels."""
        entries: list[DiffEntry] = []

        source_labels = source.metadata.labels
        target_labels = target.metadata.labels

        all_keys = set(source_labels.keys()) | set(target_labels.keys())

        for key in sorted(all_keys):
            source_val = source_labels.get(key)
            target_val = target_labels.get(key)

            if source_val is None and target_val is not None:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.METADATA,
                        change_type=ChangeType.ADDED,
                        path=f"labels/{key}",
                        old_value=None,
                        new_value=target_val,
                        severity=Severity.INFO,
                        description=f"Label {key} added",
                    )
                )
            elif source_val is not None and target_val is None:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.METADATA,
                        change_type=ChangeType.REMOVED,
                        path=f"labels/{key}",
                        old_value=source_val,
                        new_value=None,
                        severity=Severity.INFO,
                        description=f"Label {key} removed",
                    )
                )
            elif source_val != target_val:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.METADATA,
                        change_type=ChangeType.MODIFIED,
                        path=f"labels/{key}",
                        old_value=source_val,
                        new_value=target_val,
                        severity=Severity.INFO,
                        description=f"Label {key} changed",
                    )
                )

        return entries

    def _diff_layers(
        self, source: "NIMImage", target: "NIMImage"
    ) -> tuple[list[DiffEntry], list[BreakingChange]]:
        """Compare image layers."""
        entries: list[DiffEntry] = []
        breaking: list[BreakingChange] = []

        source_layers = source.metadata.manifest.layers if source.metadata.manifest else []
        target_layers = target.metadata.manifest.layers if target.metadata.manifest else []

        source_digests = {str(layer.digest) for layer in source_layers}
        target_digests = {str(layer.digest) for layer in target_layers}

        # Layers removed
        for digest in source_digests - target_digests:
            entries.append(
                DiffEntry(
                    category=ChangeCategory.LAYER,
                    change_type=ChangeType.REMOVED,
                    path=f"layers/{digest[:12]}",
                    old_value=digest,
                    new_value=None,
                    severity=Severity.INFO,
                    description="Layer removed",
                )
            )

        # Layers added
        for digest in target_digests - source_digests:
            entries.append(
                DiffEntry(
                    category=ChangeCategory.LAYER,
                    change_type=ChangeType.ADDED,
                    path=f"layers/{digest[:12]}",
                    old_value=None,
                    new_value=digest,
                    severity=Severity.INFO,
                    description="Layer added",
                )
            )

        # Layer count change
        if len(source_layers) != len(target_layers):
            entries.append(
                DiffEntry(
                    category=ChangeCategory.LAYER,
                    change_type=ChangeType.MODIFIED,
                    path="layers/count",
                    old_value=str(len(source_layers)),
                    new_value=str(len(target_layers)),
                    severity=Severity.INFO,
                    description="Number of layers changed",
                )
            )

        return entries, breaking

    def _diff_extracted_data(
        self,
        source_data: dict,
        target_data: dict,
        extractor_name: str,
    ) -> list[DiffEntry]:
        """Compare data extracted by an extractor."""
        entries: list[DiffEntry] = []

        all_keys = set(source_data.keys()) | set(target_data.keys())

        for key in sorted(all_keys):
            source_val = source_data.get(key)
            target_val = target_data.get(key)

            if source_val is None and target_val is not None:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.RUNTIME,
                        change_type=ChangeType.ADDED,
                        path=f"{extractor_name}/{key}",
                        old_value=None,
                        new_value=str(target_val),
                        severity=Severity.INFO,
                        description=f"{key} added",
                    )
                )
            elif source_val is not None and target_val is None:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.RUNTIME,
                        change_type=ChangeType.REMOVED,
                        path=f"{extractor_name}/{key}",
                        old_value=str(source_val),
                        new_value=None,
                        severity=Severity.WARNING,
                        description=f"{key} removed",
                    )
                )
            elif source_val != target_val:
                entries.append(
                    DiffEntry(
                        category=ChangeCategory.RUNTIME,
                        change_type=ChangeType.MODIFIED,
                        path=f"{extractor_name}/{key}",
                        old_value=str(source_val),
                        new_value=str(target_val),
                        severity=Severity.INFO,
                        description=f"{key} changed",
                    )
                )

        return entries



"""Optimized profiles for NIM deployments."""

from typing import Any


def get_profiles() -> dict[str, dict[str, Any]]:
    """Get optimized configuration profiles.

    Returns:
        Dictionary mapping profile names to their configurations
    """
    return {
        "high-throughput": {
            "name": "High Throughput",
            "description": "Optimized for maximum requests per second",
            "use_case": "Production deployments with high concurrency",
            "env": {
                "NIM_MAX_BATCH_SIZE": "32",
                "NIM_MAX_CONCURRENT_REQUESTS": "64",
                "NIM_GPU_MEMORY_UTILIZATION": "0.95",
                "NIM_ENABLE_CHUNKED_PREFILL": "true",
                "NIM_ENABLE_PREFIX_CACHING": "true",
            },
            "requirements": {
                "min_memory_gb": 40,
                "recommended_gpus": ["H100", "A100-80GB"],
            },
        },
        "low-latency": {
            "name": "Low Latency",
            "description": "Optimized for fastest time-to-first-token",
            "use_case": "Interactive applications requiring fast responses",
            "env": {
                "NIM_MAX_BATCH_SIZE": "1",
                "NIM_MAX_CONCURRENT_REQUESTS": "8",
                "NIM_GPU_MEMORY_UTILIZATION": "0.85",
                "NIM_ENABLE_CHUNKED_PREFILL": "false",
            },
            "requirements": {
                "min_memory_gb": 24,
                "recommended_gpus": ["H100", "L40S", "A10"],
            },
        },
        "memory-efficient": {
            "name": "Memory Efficient",
            "description": "Optimized for running larger models on smaller GPUs",
            "use_case": "Resource-constrained environments",
            "env": {
                "NIM_MAX_BATCH_SIZE": "4",
                "NIM_MAX_CONCURRENT_REQUESTS": "8",
                "NIM_GPU_MEMORY_UTILIZATION": "0.9",
                "NIM_QUANTIZATION": "int8",
                "NIM_KV_CACHE_DTYPE": "fp8",
                "NIM_SWAP_SPACE": "4",
            },
            "requirements": {
                "min_memory_gb": 16,
                "recommended_gpus": ["L4", "T4", "A10"],
            },
        },
        "balanced": {
            "name": "Balanced",
            "description": "Balanced configuration for general use",
            "use_case": "Default production configuration",
            "env": {
                "NIM_MAX_BATCH_SIZE": "8",
                "NIM_MAX_CONCURRENT_REQUESTS": "16",
                "NIM_GPU_MEMORY_UTILIZATION": "0.9",
                "NIM_ENABLE_CHUNKED_PREFILL": "true",
            },
            "requirements": {
                "min_memory_gb": 24,
                "recommended_gpus": ["H100", "A100", "L40", "A10"],
            },
        },
        "development": {
            "name": "Development",
            "description": "Configuration for development and testing",
            "use_case": "Local development and debugging",
            "env": {
                "NIM_MAX_BATCH_SIZE": "1",
                "NIM_MAX_CONCURRENT_REQUESTS": "4",
                "NIM_GPU_MEMORY_UTILIZATION": "0.7",
                "NIM_LOG_LEVEL": "DEBUG",
                "NIM_METRICS_ENABLED": "true",
            },
            "requirements": {
                "min_memory_gb": 8,
                "recommended_gpus": ["RTX 4090", "RTX 3090", "T4"],
            },
        },
        "multi-gpu": {
            "name": "Multi-GPU",
            "description": "Configuration for multi-GPU tensor parallelism",
            "use_case": "Large model inference across multiple GPUs",
            "env": {
                "NIM_TENSOR_PARALLEL_SIZE": "2",
                "NIM_MAX_BATCH_SIZE": "16",
                "NIM_MAX_CONCURRENT_REQUESTS": "32",
                "NIM_GPU_MEMORY_UTILIZATION": "0.9",
            },
            "requirements": {
                "min_gpus": 2,
                "min_memory_gb": 48,
                "recommended_gpus": ["H100", "A100-80GB"],
            },
        },
        "cost-optimized": {
            "name": "Cost Optimized",
            "description": "Configuration optimized for cloud cost efficiency",
            "use_case": "Cloud deployments with cost constraints",
            "env": {
                "NIM_MAX_BATCH_SIZE": "16",
                "NIM_MAX_CONCURRENT_REQUESTS": "32",
                "NIM_GPU_MEMORY_UTILIZATION": "0.95",
                "NIM_QUANTIZATION": "int8",
                "NIM_ENABLE_PREFIX_CACHING": "true",
            },
            "requirements": {
                "min_memory_gb": 16,
                "recommended_gpus": ["L4", "T4", "A10G"],
            },
        },
    }


def get_profile(name: str) -> dict[str, Any] | None:
    """Get a specific profile by name.

    Args:
        name: Profile name

    Returns:
        Profile configuration or None if not found
    """
    return get_profiles().get(name)


def suggest_profile(
    model_size_gb: float,
    gpu_memory_gb: float,
    priority: str = "balanced",
) -> str:
    """Suggest a profile based on requirements.

    Args:
        model_size_gb: Model size in GB
        gpu_memory_gb: Available GPU memory in GB
        priority: Priority ("throughput", "latency", "memory", "balanced")

    Returns:
        Suggested profile name
    """
    memory_ratio = model_size_gb / gpu_memory_gb

    if memory_ratio > 0.8:
        return "memory-efficient"

    if priority == "throughput":
        return "high-throughput"
    elif priority == "latency":
        return "low-latency"
    elif priority == "memory":
        return "memory-efficient"
    else:
        return "balanced"



"""Utility functions for nim-audit."""

from nim_audit.utils.hashing import compute_hash, hash_dict, hash_file, short_hash
from nim_audit.utils.logging import configure_logging, get_logger, get_logger_with_context
from nim_audit.utils.errors import (
    NimAuditError,
    ImageNotFoundError,
    AuthenticationError,
    ValidationError,
    ConfigurationError,
    NetworkError,
    TimeoutError,
    retry,
    validate_image_reference,
    validate_env_var_name,
    safe_get,
)
from nim_audit.utils.expression import SafeExpressionEvaluator, safe_eval
from nim_audit.utils.cache import Cache, get_cache
from nim_audit.utils.config import (
    NimAuditConfig,
    CacheConfig,
    RegistryConfig,
    OutputConfig,
    LintConfig,
    load_config,
    save_config,
    get_config,
    set_config,
)
from nim_audit.utils.plugins import (
    Plugin,
    PluginContext,
    PluginManager,
    get_plugin_manager,
)

__all__ = [
    # Hashing
    "compute_hash",
    "hash_dict",
    "hash_file",
    "short_hash",
    # Logging
    "configure_logging",
    "get_logger",
    "get_logger_with_context",
    # Errors
    "NimAuditError",
    "ImageNotFoundError",
    "AuthenticationError",
    "ValidationError",
    "ConfigurationError",
    "NetworkError",
    "TimeoutError",
    "retry",
    "validate_image_reference",
    "validate_env_var_name",
    "safe_get",
    # Expression
    "SafeExpressionEvaluator",
    "safe_eval",
    # Cache
    "Cache",
    "get_cache",
    # Config
    "NimAuditConfig",
    "CacheConfig",
    "RegistryConfig",
    "OutputConfig",
    "LintConfig",
    "load_config",
    "save_config",
    "get_config",
    "set_config",
    # Plugins
    "Plugin",
    "PluginContext",
    "PluginManager",
    "get_plugin_manager",
]



"""OCI registry client implementation."""

from __future__ import annotations

import hashlib
import json
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import httpx

from nim_audit.models.image import ImageDigest, ImageManifest, ImageMetadata, LayerInfo
from nim_audit.registry.base import (
    Registry,
    RegistryAuth,
    RegistryAuthError,
    RegistryError,
    RegistryNotFoundError,
)


class OCIRegistry:
    """Registry client for OCI-compliant container registries.

    Implements the OCI Distribution Specification for interacting
    with container registries like Docker Hub, GHCR, ECR, etc.

    Example:
        registry = OCIRegistry("https://registry-1.docker.io")
        metadata = registry.get_metadata("library/nginx:latest")
    """

    # Well-known registry URLs
    REGISTRY_URLS = {
        "docker.io": "https://registry-1.docker.io",
        "registry-1.docker.io": "https://registry-1.docker.io",
        "ghcr.io": "https://ghcr.io",
        "gcr.io": "https://gcr.io",
        "quay.io": "https://quay.io",
        "nvcr.io": "https://nvcr.io",
    }

    # Media types
    MANIFEST_V2 = "application/vnd.docker.distribution.manifest.v2+json"
    MANIFEST_LIST = "application/vnd.docker.distribution.manifest.list.v2+json"
    OCI_MANIFEST = "application/vnd.oci.image.manifest.v1+json"
    OCI_INDEX = "application/vnd.oci.image.index.v1+json"

    def __init__(
        self,
        base_url: str | None = None,
        auth: RegistryAuth | None = None,
        timeout: float = 30.0,
        max_retries: int = 3,
    ) -> None:
        """Initialize the OCI registry client.

        Args:
            base_url: Base URL of the registry (auto-detected from reference if None)
            auth: Authentication credentials
            timeout: Request timeout in seconds
            max_retries: Maximum number of retry attempts
        """
        self._base_url = base_url
        self._auth = auth or RegistryAuth.from_env()
        self._timeout = timeout
        self._max_retries = max_retries
        self._token_cache: dict[str, str] = {}

    def _get_registry_url(self, registry: str | None) -> str:
        """Get the registry URL for a registry hostname."""
        if not registry:
            return self.REGISTRY_URLS["docker.io"]

        if registry in self.REGISTRY_URLS:
            return self.REGISTRY_URLS[registry]

        # Assume HTTPS for unknown registries
        if not registry.startswith("http"):
            return f"https://{registry}"
        return registry

    def _get_client(self) -> httpx.Client:
        """Create an HTTP client with retry support."""
        transport = httpx.HTTPTransport(retries=self._max_retries)
        return httpx.Client(
            timeout=self._timeout,
            transport=transport,
            follow_redirects=True,
        )

    def _get_token(self, client: httpx.Client, www_authenticate: str, repository: str) -> str:
        """Get a bearer token for authentication.

        Args:
            client: HTTP client
            www_authenticate: WWW-Authenticate header value
            repository: Repository name for scope

        Returns:
            Bearer token
        """
        # Parse WWW-Authenticate header
        # Format: Bearer realm="...",service="...",scope="..."
        params = {}
        for part in www_authenticate.replace("Bearer ", "").split(","):
            if "=" in part:
                key, value = part.split("=", 1)
                params[key.strip()] = value.strip().strip('"')

        realm = params.get("realm")
        if not realm:
            raise RegistryAuthError("No realm in WWW-Authenticate header")

        # Build token request
        token_params = {
            "service": params.get("service", ""),
            "scope": f"repository:{repository}:pull",
        }

        auth = None
        if self._auth:
            if self._auth.token:
                # Use token directly
                return self._auth.token
            elif self._auth.username and self._auth.password:
                auth = (self._auth.username, self._auth.password)

        response = client.get(realm, params=token_params, auth=auth)

        if response.status_code == 401:
            raise RegistryAuthError("Token authentication failed")
        elif response.status_code != 200:
            raise RegistryError(f"Token request failed: {response.status_code}")

        data = response.json()
        return data.get("token") or data.get("access_token", "")

    def _request(
        self,
        client: httpx.Client,
        method: str,
        url: str,
        repository: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> httpx.Response:
        """Make an authenticated request to the registry.

        Args:
            client: HTTP client
            method: HTTP method
            url: Request URL
            repository: Repository name (for token scope)
            headers: Additional headers
            **kwargs: Additional request arguments

        Returns:
            HTTP response
        """
        headers = headers or {}

        # Try with cached token first
        cache_key = f"{url}:{repository}"
        if cache_key in self._token_cache:
            headers["Authorization"] = f"Bearer {self._token_cache[cache_key]}"

        response = client.request(method, url, headers=headers, **kwargs)

        # Handle 401 - need to authenticate
        if response.status_code == 401:
            www_auth = response.headers.get("www-authenticate", "")
            if "bearer" in www_auth.lower():
                token = self._get_token(client, www_auth, repository)
                self._token_cache[cache_key] = token
                headers["Authorization"] = f"Bearer {token}"
                response = client.request(method, url, headers=headers, **kwargs)

        return response

    def get_manifest(self, reference: str) -> ImageManifest:
        """Get the manifest for an image.

        Args:
            reference: Image reference (e.g., "nginx:latest")

        Returns:
            The image manifest
        """
        parsed = self._parse_reference(reference)
        registry_url = self._get_registry_url(parsed["registry"])
        repository = parsed["repository"]
        tag = parsed["tag"] or parsed["digest"] or "latest"

        # Handle Docker Hub library images
        if "docker.io" in registry_url and "/" not in repository:
            repository = f"library/{repository}"

        url = f"{registry_url}/v2/{repository}/manifests/{tag}"

        with self._get_client() as client:
            headers = {
                "Accept": ", ".join(
                    [self.MANIFEST_V2, self.OCI_MANIFEST, self.MANIFEST_LIST, self.OCI_INDEX]
                )
            }

            response = self._request(client, "GET", url, repository, headers=headers)

            if response.status_code == 404:
                raise RegistryNotFoundError(reference)
            elif response.status_code == 401:
                raise RegistryAuthError(f"Authentication failed for {reference}")
            elif response.status_code != 200:
                raise RegistryError(f"Failed to get manifest: {response.status_code}")

            data = response.json()
            content_type = response.headers.get("content-type", "")

            # Handle manifest list/index - get first amd64/linux manifest
            if "list" in content_type or "index" in content_type:
                manifests = data.get("manifests", [])
                for m in manifests:
                    platform = m.get("platform", {})
                    if platform.get("architecture") == "amd64" and platform.get("os") == "linux":
                        # Fetch the actual manifest
                        digest = m.get("digest")
                        url = f"{registry_url}/v2/{repository}/manifests/{digest}"
                        response = self._request(
                            client, "GET", url, repository, headers={"Accept": self.MANIFEST_V2}
                        )
                        data = response.json()
                        break

            # Parse manifest
            layers = []
            for layer in data.get("layers", []):
                layers.append(
                    LayerInfo(
                        digest=ImageDigest.from_string(layer.get("digest", "")),
                        size=layer.get("size", 0),
                        media_type=layer.get("mediaType", ""),
                    )
                )

            # Calculate manifest digest
            manifest_bytes = response.content
            manifest_digest = hashlib.sha256(manifest_bytes).hexdigest()

            config_digest_str = data.get("config", {}).get("digest", "")

            return ImageManifest(
                schema_version=data.get("schemaVersion", 2),
                media_type=data.get("mediaType", self.MANIFEST_V2),
                digest=ImageDigest(hash=manifest_digest),
                config_digest=ImageDigest.from_string(config_digest_str)
                if config_digest_str
                else ImageDigest(hash=""),
                layers=layers,
            )

    def get_metadata(self, reference: str) -> ImageMetadata:
        """Get comprehensive metadata for an image.

        Args:
            reference: Image reference

        Returns:
            The image metadata
        """
        parsed = self._parse_reference(reference)
        registry_url = self._get_registry_url(parsed["registry"])
        repository = parsed["repository"]

        # Handle Docker Hub library images
        if "docker.io" in registry_url and "/" not in repository:
            repository = f"library/{repository}"

        # Get manifest first
        manifest = self.get_manifest(reference)

        # Fetch config blob
        config_url = f"{registry_url}/v2/{repository}/blobs/{manifest.config_digest}"

        with self._get_client() as client:
            response = self._request(client, "GET", config_url, repository)

            if response.status_code != 200:
                raise RegistryError(f"Failed to get config blob: {response.status_code}")

            config = response.json()

        # Parse config
        container_config = config.get("config", {})
        labels = container_config.get("Labels", {}) or {}

        # Parse environment
        env_list = container_config.get("Env", []) or []
        env = {}
        for item in env_list:
            if "=" in item:
                key, value = item.split("=", 1)
                env[key] = value

        # Parse exposed ports
        exposed_ports = []
        ports_config = container_config.get("ExposedPorts", {}) or {}
        for port_spec in ports_config.keys():
            import re

            port_match = re.match(r"(\d+)", port_spec)
            if port_match:
                exposed_ports.append(int(port_match.group(1)))

        # Parse creation timestamp
        created = None
        created_str = config.get("created")
        if created_str:
            try:
                from datetime import datetime

                created_str = created_str.replace("Z", "+00:00")
                created = datetime.fromisoformat(created_str[:26] + created_str[-6:])
            except (ValueError, TypeError):
                pass

        # Extract NIM-specific metadata
        nim_version = labels.get("com.nvidia.nim.version")
        model_name = labels.get("com.nvidia.nim.model.name")
        model_version = labels.get("com.nvidia.nim.model.version")
        quantization = labels.get("com.nvidia.nim.model.quantization")

        return ImageMetadata(
            reference=reference,
            repository=parsed["repository"] or reference,
            tag=parsed["tag"],
            digest=manifest.digest,
            manifest=manifest,
            labels=labels,
            created=created,
            architecture=config.get("architecture"),
            os=config.get("os"),
            nim_version=nim_version,
            model_name=model_name,
            model_version=model_version,
            quantization=quantization,
            env=env,
            exposed_ports=exposed_ports,
            entrypoint=container_config.get("Entrypoint", []) or [],
            cmd=container_config.get("Cmd", []) or [],
            raw_config=config,
        )

    def pull_layer(self, reference: str, digest: str, dest: Path) -> None:
        """Download a layer blob to the specified destination.

        Args:
            reference: Image reference
            digest: Layer digest
            dest: Destination path
        """
        parsed = self._parse_reference(reference)
        registry_url = self._get_registry_url(parsed["registry"])
        repository = parsed["repository"]

        if "docker.io" in registry_url and "/" not in repository:
            repository = f"library/{repository}"

        url = f"{registry_url}/v2/{repository}/blobs/{digest}"

        with self._get_client() as client:
            response = self._request(client, "GET", url, repository)

            if response.status_code == 404:
                raise RegistryNotFoundError(f"Layer {digest}")
            elif response.status_code != 200:
                raise RegistryError(f"Failed to pull layer: {response.status_code}")

            dest.write_bytes(response.content)

    def list_tags(self, repository: str) -> list[str]:
        """List all tags for a repository.

        Args:
            repository: Repository name

        Returns:
            List of tag names
        """
        parsed = self._parse_reference(repository)
        registry_url = self._get_registry_url(parsed["registry"])
        repo = parsed["repository"] or repository

        if "docker.io" in registry_url and "/" not in repo:
            repo = f"library/{repo}"

        url = f"{registry_url}/v2/{repo}/tags/list"

        with self._get_client() as client:
            response = self._request(client, "GET", url, repo)

            if response.status_code == 404:
                raise RegistryNotFoundError(repository)
            elif response.status_code != 200:
                raise RegistryError(f"Failed to list tags: {response.status_code}")

            data = response.json()
            return sorted(data.get("tags", []))

    @staticmethod
    def _parse_reference(reference: str) -> dict[str, str | None]:
        """Parse an image reference into components."""
        result: dict[str, str | None] = {
            "registry": None,
            "repository": None,
            "tag": None,
            "digest": None,
        }

        # Handle digest
        if "@" in reference:
            ref_part, digest = reference.rsplit("@", 1)
            result["digest"] = digest
            reference = ref_part

        # Handle tag
        if ":" in reference:
            parts = reference.rsplit(":", 1)
            if "/" not in parts[1] and not parts[1].isdigit():
                reference = parts[0]
                result["tag"] = parts[1]

        # Handle registry and repository
        parts = reference.split("/")
        if len(parts) == 1:
            result["repository"] = parts[0]
        elif len(parts) == 2:
            if "." in parts[0] or ":" in parts[0] or parts[0] == "localhost":
                result["registry"] = parts[0]
                result["repository"] = parts[1]
            else:
                result["repository"] = reference
        else:
            result["registry"] = parts[0]
            result["repository"] = "/".join(parts[1:])

        return result



"""Shared utilities for CLI commands."""

from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

import typer
from pydantic import BaseModel
from rich.console import Console

if TYPE_CHECKING:
    from nim_audit.core.image import NIMImage

T = TypeVar("T", bound=BaseModel)

# Shared console instance
console = Console()


def load_image(image: str) -> "NIMImage":
    """Load a NIM image from local or registry.

    Tries local first, falls back to registry.

    Args:
        image: Image reference

    Returns:
        Loaded NIMImage instance
    """
    from nim_audit.core.image import NIMImage

    with console.status("Loading image..."):
        try:
            return NIMImage.from_local(image)
        except ValueError:
            return NIMImage.from_registry(image)


def handle_result_errors(result: Any, error_message: str = "Operation failed") -> None:
    """Handle errors in a result object and exit if failed.

    Args:
        result: Result object with success and errors attributes
        error_message: Message to display on error
    """
    if not result.success:
        console.print(f"[red]Error:[/red] {error_message}")
        for error in result.errors:
            console.print(f"  {error}")
        raise typer.Exit(1)


def check_report(report: Any | None, error_message: str = "No report generated") -> None:
    """Check that a report was generated and exit if not.

    Args:
        report: Report object (may be None)
        error_message: Message to display if report is None
    """
    if report is None:
        console.print(f"[red]Error:[/red] {error_message}")
        raise typer.Exit(1)


def output_json(data: dict[str, Any] | BaseModel, output: Path | None = None) -> None:
    """Output data as JSON to console or file.

    Args:
        data: Data to output (dict or Pydantic model)
        output: Optional output file path
    """
    if isinstance(data, BaseModel):
        data = data.model_dump(mode="json")

    json_str = json.dumps(data, indent=2, default=str)

    if output:
        output.write_text(json_str)
        console.print(f"Report written to {output}")
    else:
        console.print(json_str)


def severity_style(severity: str) -> str:
    """Get Rich style for a severity level.

    Args:
        severity: Severity value (error, warning, info)

    Returns:
        Rich style string
    """
    styles = {
        "error": "red",
        "warning": "yellow",
        "info": "blue",
        "critical": "bold red",
        "high": "red",
        "medium": "yellow",
        "low": "green",
        "breaking": "bold red",
    }
    return styles.get(severity.lower(), "white")


def status_icon(success: bool) -> str:
    """Get a colored status icon.

    Args:
        success: Whether the status is successful

    Returns:
        Formatted status string
    """
    return "[green]OK[/green]" if success else "[red]FAIL[/red]"



"""NIMImage class for container inspection."""

from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path
from typing import Any

from nim_audit.models.common import AuditError
from nim_audit.models.image import ImageDigest, ImageManifest, ImageMetadata, LayerInfo


class NIMImage:
    """Represents a NIM container image for inspection.

    This class provides methods to load and inspect NIM container images
    from various sources (local Docker daemon, remote registries).

    Example:
        # From registry
        image = NIMImage.from_registry("nvcr.io/nim/llama3:1.5.0")

        # From local Docker
        image = NIMImage.from_local("nvcr.io/nim/llama3:1.5.0")

        # Access metadata
        print(image.metadata.nim_version)
        print(image.metadata.model_name)
    """

    def __init__(self, metadata: ImageMetadata) -> None:
        """Initialize with image metadata.

        Args:
            metadata: The image metadata
        """
        self._metadata = metadata
        self._extracted_fs: Path | None = None

    @property
    def metadata(self) -> ImageMetadata:
        """Get the image metadata."""
        return self._metadata

    @property
    def reference(self) -> str:
        """Get the image reference."""
        return self._metadata.reference

    @property
    def tag(self) -> str | None:
        """Get the image tag."""
        return self._metadata.tag

    @property
    def digest(self) -> ImageDigest | None:
        """Get the image digest."""
        return self._metadata.digest

    @classmethod
    def from_registry(
        cls,
        reference: str,
        auth: dict[str, str] | None = None,
    ) -> "NIMImage":
        """Load an image from a container registry.

        Args:
            reference: Image reference (e.g., "nvcr.io/nim/llama3:1.5.0")
            auth: Optional authentication credentials

        Returns:
            NIMImage instance

        Raises:
            ValueError: If the reference is invalid
        """
        parsed = cls._parse_reference(reference)
        metadata = cls._fetch_registry_metadata(reference, auth)
        return cls(metadata)

    @classmethod
    def from_local(cls, reference: str) -> "NIMImage":
        """Load an image from the local Docker daemon.

        Args:
            reference: Image reference or ID

        Returns:
            NIMImage instance

        Raises:
            ValueError: If the image is not found locally
        """
        metadata = cls._fetch_local_metadata(reference)
        return cls(metadata)

    @classmethod
    def from_metadata(cls, metadata: ImageMetadata) -> "NIMImage":
        """Create an image from existing metadata.

        Args:
            metadata: Pre-existing image metadata

        Returns:
            NIMImage instance
        """
        return cls(metadata)

    @staticmethod
    def _parse_reference(reference: str) -> dict[str, str | None]:
        """Parse an image reference into components.

        Args:
            reference: Image reference string

        Returns:
            Dict with registry, repository, tag, and digest
        """
        result: dict[str, str | None] = {
            "registry": None,
            "repository": None,
            "tag": None,
            "digest": None,
        }

        # Handle digest
        if "@" in reference:
            ref_part, digest = reference.rsplit("@", 1)
            result["digest"] = digest
            reference = ref_part

        # Handle tag
        if ":" in reference:
            # Check if it's a port or a tag
            parts = reference.rsplit(":", 1)
            if "/" in parts[1] or parts[1].isdigit():
                # It's a port number, not a tag
                pass
            else:
                reference = parts[0]
                result["tag"] = parts[1]

        # Handle registry and repository
        parts = reference.split("/")
        if len(parts) == 1:
            # Just a repository name (e.g., "ubuntu")
            result["repository"] = parts[0]
        elif len(parts) == 2:
            # Could be registry/repo or namespace/repo
            if "." in parts[0] or ":" in parts[0] or parts[0] == "localhost":
                result["registry"] = parts[0]
                result["repository"] = parts[1]
            else:
                result["repository"] = reference
        else:
            # registry/namespace/repo or registry/namespace/.../repo
            result["registry"] = parts[0]
            result["repository"] = "/".join(parts[1:])

        return result

    @staticmethod
    def _fetch_registry_metadata(
        reference: str,
        auth: dict[str, str] | None = None,
    ) -> ImageMetadata:
        """Fetch metadata from a container registry.

        This is a placeholder implementation. In production, this would
        use the OCI registry API or Docker registry API.
        """
        parsed = NIMImage._parse_reference(reference)

        # For now, return placeholder metadata
        # Real implementation would fetch from registry
        return ImageMetadata(
            reference=reference,
            repository=parsed.get("repository") or reference,
            tag=parsed.get("tag"),
            digest=ImageDigest.from_string(parsed["digest"]) if parsed.get("digest") else None,
            labels={},
            env={},
        )

    @staticmethod
    def _fetch_local_metadata(reference: str) -> ImageMetadata:
        """Fetch metadata from local Docker daemon.

        This implementation uses the Docker SDK to inspect images.
        """
        try:
            import docker

            client = docker.from_env()
            image = client.images.get(reference)
            config = image.attrs

            # Parse labels
            labels = config.get("Config", {}).get("Labels", {}) or {}

            # Parse environment variables
            env_list = config.get("Config", {}).get("Env", []) or []
            env = {}
            for item in env_list:
                if "=" in item:
                    key, value = item.split("=", 1)
                    env[key] = value

            # Parse exposed ports
            exposed_ports = []
            ports_config = config.get("Config", {}).get("ExposedPorts", {}) or {}
            for port_spec in ports_config.keys():
                port_match = re.match(r"(\d+)", port_spec)
                if port_match:
                    exposed_ports.append(int(port_match.group(1)))

            # Parse creation timestamp
            created = None
            created_str = config.get("Created")
            if created_str:
                try:
                    # Handle both formats: with and without fractional seconds
                    created_str = created_str.replace("Z", "+00:00")
                    if "." in created_str:
                        # Truncate fractional seconds to 6 digits
                        parts = created_str.split(".")
                        frac_and_tz = parts[1]
                        # Find where timezone starts
                        tz_start = -1
                        for i, c in enumerate(frac_and_tz):
                            if c in "+-":
                                tz_start = i
                                break
                        if tz_start > 0:
                            frac = frac_and_tz[:tz_start][:6]
                            tz = frac_and_tz[tz_start:]
                            created_str = f"{parts[0]}.{frac}{tz}"
                    created = datetime.fromisoformat(created_str)
                except (ValueError, TypeError):
                    pass

            # Build manifest from layer info
            layers = []
            for layer_digest in image.attrs.get("RootFS", {}).get("Layers", []):
                layers.append(
                    LayerInfo(
                        digest=ImageDigest.from_string(layer_digest),
                        size=0,  # Size not available from this API
                        media_type="application/vnd.docker.image.rootfs.diff.tar.gzip",
                    )
                )

            manifest = None
            if layers:
                manifest = ImageManifest(
                    schema_version=2,
                    media_type="application/vnd.docker.distribution.manifest.v2+json",
                    digest=ImageDigest(hash=image.id.replace("sha256:", "")),
                    config_digest=ImageDigest(hash=image.id.replace("sha256:", "")),
                    layers=layers,
                )

            # Extract NIM-specific metadata from labels
            nim_version = labels.get("com.nvidia.nim.version")
            model_name = labels.get("com.nvidia.nim.model.name")
            model_version = labels.get("com.nvidia.nim.model.version")
            quantization = labels.get("com.nvidia.nim.model.quantization")

            parsed = NIMImage._parse_reference(reference)

            return ImageMetadata(
                reference=reference,
                repository=parsed.get("repository") or reference,
                tag=parsed.get("tag"),
                digest=ImageDigest(hash=image.id.replace("sha256:", "")),
                manifest=manifest,
                labels=labels,
                created=created,
                architecture=config.get("Architecture"),
                os=config.get("Os"),
                nim_version=nim_version,
                model_name=model_name,
                model_version=model_version,
                quantization=quantization,
                env=env,
                exposed_ports=exposed_ports,
                entrypoint=config.get("Config", {}).get("Entrypoint", []) or [],
                cmd=config.get("Config", {}).get("Cmd", []) or [],
                raw_config=config,
            )

        except ImportError:
            raise RuntimeError("Docker SDK not available. Install with: pip install docker")
        except Exception as e:
            raise ValueError(f"Failed to load image '{reference}': {e}")

    def __repr__(self) -> str:
        return f"NIMImage(reference='{self.reference}')"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, NIMImage):
            return NotImplemented
        # Compare by digest if available, otherwise by reference
        if self.digest and other.digest:
            return self.digest == other.digest
        return self.reference == other.reference



"""CLI command for diffing NIM containers."""

from pathlib import Path
from typing import Optional

import typer
from rich.console import Console
from rich.table import Table
from rich.panel import Panel

from nim_audit.models.diff import Severity

app = typer.Typer(help="Compare two NIM container versions.")
console = Console()


@app.callback(invoke_without_command=True)
def diff_cmd(
    ctx: typer.Context,
    source: str = typer.Argument(..., help="Source (old) image reference"),
    target: str = typer.Argument(..., help="Target (new) image reference"),
    format: str = typer.Option(
        "terminal",
        "--format",
        "-f",
        help="Output format (terminal, json, markdown)",
    ),
    output: Optional[Path] = typer.Option(
        None,
        "--output",
        "-o",
        help="Output file path",
    ),
    breaking_only: bool = typer.Option(
        False,
        "--breaking-only",
        help="Only show breaking changes",
    ),
    category: Optional[str] = typer.Option(
        None,
        "--category",
        "-c",
        help="Filter by category (metadata, model, tokenizer, api, runtime, layer, config, environment)",
    ),
) -> None:
    """
    Compare two NIM container versions.

    Analyzes differences in metadata, model artifacts, configuration,
    and API schemas between two container images.

    Example:
        nim-audit diff nvcr.io/nim/llama3:1.5.0 nvcr.io/nim/llama3:1.6.0
    """
    from nim_audit.core.image import NIMImage
    from nim_audit.core.diff import DiffEngine
    from nim_audit.models.diff import ChangeCategory

    with console.status(f"Loading images..."):
        try:
            source_image = NIMImage.from_local(source)
        except ValueError:
            source_image = NIMImage.from_registry(source)

        try:
            target_image = NIMImage.from_local(target)
        except ValueError:
            target_image = NIMImage.from_registry(target)

    with console.status("Analyzing differences..."):
        engine = DiffEngine()
        result = engine.diff(source_image, target_image)

    if not result.success:
        console.print("[red]Error:[/red] Failed to diff images")
        for error in result.errors:
            console.print(f"  {error}")
        raise typer.Exit(1)

    report = result.report
    if report is None:
        console.print("[red]Error:[/red] No report generated")
        raise typer.Exit(1)

    # Filter entries if category specified
    entries = report.entries
    if category:
        try:
            cat = ChangeCategory(category)
            entries = [e for e in entries if e.category == cat]
        except ValueError:
            console.print(f"[red]Error:[/red] Invalid category: {category}")
            raise typer.Exit(1)

    if format == "json":
        import json

        output_data = report.model_dump(mode="json")
        json_str = json.dumps(output_data, indent=2, default=str)

        if output:
            output.write_text(json_str)
            console.print(f"Report written to {output}")
        else:
            console.print(json_str)

    elif format == "markdown":
        md = _generate_markdown(report)
        if output:
            output.write_text(md)
            console.print(f"Report written to {output}")
        else:
            console.print(md)

    else:  # terminal
        _print_terminal_report(report, entries, breaking_only)


def _print_terminal_report(report, entries, breaking_only: bool) -> None:
    """Print a rich terminal report."""
    # Header
    console.print()
    console.print(
        Panel(
            f"[bold]Source:[/bold] {report.source_image.reference}\n"
            f"[bold]Target:[/bold] {report.target_image.reference}",
            title="NIM Diff Report",
        )
    )

    # Breaking changes
    if report.breaking_changes:
        console.print()
        console.print("[bold red]Breaking Changes[/bold red]")
        for bc in report.breaking_changes:
            console.print(f"  [red]![/red] {bc.title}")
            console.print(f"      {bc.description}")
            if bc.migration:
                console.print(f"      [dim]Migration: {bc.migration}[/dim]")

    if breaking_only:
        return

    # Summary
    console.print()
    table = Table(title="Summary", show_header=False)
    table.add_column("Metric", style="bold")
    table.add_column("Count")
    table.add_row("Total Changes", str(report.total_changes))
    table.add_row("Added", f"[green]{report.added_count}[/green]")
    table.add_row("Removed", f"[red]{report.removed_count}[/red]")
    table.add_row("Modified", f"[yellow]{report.modified_count}[/yellow]")
    table.add_row("Breaking", f"[bold red]{len(report.breaking_changes)}[/bold red]")
    console.print(table)

    # Detailed changes
    if entries:
        console.print()
        table = Table(title="Changes")
        table.add_column("Category", style="dim")
        table.add_column("Type")
        table.add_column("Path")
        table.add_column("Old Value", max_width=30)
        table.add_column("New Value", max_width=30)

        for entry in entries:
            type_style = {
                "added": "green",
                "removed": "red",
                "modified": "yellow",
            }.get(entry.change_type.value, "white")

            severity_marker = ""
            if entry.severity == Severity.BREAKING:
                severity_marker = " [bold red]![/bold red]"

            table.add_row(
                entry.category.value,
                f"[{type_style}]{entry.change_type.value}[/{type_style}]{severity_marker}",
                entry.path,
                entry.old_value or "-",
                entry.new_value or "-",
            )

        console.print(table)


def _generate_markdown(report) -> str:
    """Generate a markdown report."""
    lines = [
        "# NIM Diff Report",
        "",
        f"**Source:** `{report.source_image.reference}`",
        f"**Target:** `{report.target_image.reference}`",
        f"**Generated:** {report.generated_at}",
        "",
        "## Summary",
        "",
        f"- Total Changes: {report.total_changes}",
        f"- Added: {report.added_count}",
        f"- Removed: {report.removed_count}",
        f"- Modified: {report.modified_count}",
        f"- Breaking Changes: {len(report.breaking_changes)}",
        "",
    ]

    if report.breaking_changes:
        lines.extend(
            [
                "## Breaking Changes",
                "",
            ]
        )
        for bc in report.breaking_changes:
            lines.extend(
                [
                    f"### {bc.title}",
                    "",
                    bc.description,
                    "",
                    f"**Impact:** {bc.impact}",
                    "",
                ]
            )
            if bc.migration:
                lines.append(f"**Migration:** {bc.migration}")
                lines.append("")

    lines.extend(
        [
            "## All Changes",
            "",
            "| Category | Type | Path | Old | New |",
            "|----------|------|------|-----|-----|",
        ]
    )

    for entry in report.entries:
        old = entry.old_value or "-"
        new = entry.new_value or "-"
        # Escape pipe characters
        old = old.replace("|", "\\|")[:50]
        new = new.replace("|", "\\|")[:50]
        lines.append(
            f"| {entry.category.value} | {entry.change_type.value} | "
            f"`{entry.path}` | {old} | {new} |"
        )

    return "\n".join(lines)



"""Unit tests for the ConfigAnalyzer."""

import pytest

from nim_audit.core.config import ConfigAnalyzer
from nim_audit.core.image import NIMImage
from nim_audit.models.config import ImpactLevel


class TestConfigAnalyzer:
    """Tests for ConfigAnalyzer."""

    def test_analyze_basic(self, sample_nim_image: NIMImage):
        """Test basic config analysis."""
        analyzer = ConfigAnalyzer()
        result = analyzer.analyze(sample_nim_image)

        assert result.success
        assert result.report is not None
        assert len(result.report.entries) > 0

    def test_analyze_with_env(self, sample_nim_image: NIMImage):
        """Test config analysis with custom env vars."""
        analyzer = ConfigAnalyzer()
        result = analyzer.analyze(
            sample_nim_image,
            env={"NIM_MAX_BATCH_SIZE": "64", "NIM_LOG_LEVEL": "DEBUG"},
        )

        assert result.success
        assert result.report is not None

        # Find the batch size entry
        batch_entry = next(
            (e for e in result.report.entries if e.name == "NIM_MAX_BATCH_SIZE"),
            None,
        )
        assert batch_entry is not None
        assert batch_entry.value == "64"

    def test_analyze_with_env_file(self, sample_nim_image: NIMImage, sample_env_file: str):
        """Test config analysis with env file."""
        analyzer = ConfigAnalyzer()
        result = analyzer.analyze(sample_nim_image, env_file=sample_env_file)

        assert result.success
        assert result.report is not None

        # Check that env file values are loaded
        batch_entry = next(
            (e for e in result.report.entries if e.name == "NIM_MAX_BATCH_SIZE"),
            None,
        )
        assert batch_entry is not None
        assert batch_entry.value == "32"

    def test_analyze_detects_deprecated(self, sample_nim_image: NIMImage):
        """Test that deprecated config vars are detected."""
        analyzer = ConfigAnalyzer()
        result = analyzer.analyze(
            sample_nim_image,
            env={"NIM_BATCH_SIZE": "16"},  # Deprecated
        )

        assert result.success
        assert result.report is not None

        deprecated = result.report.deprecated_entries
        assert len(deprecated) > 0
        assert any(e.name == "NIM_BATCH_SIZE" for e in deprecated)

    def test_analyze_high_impact_entries(self, sample_nim_image: NIMImage):
        """Test that high impact entries are identified."""
        analyzer = ConfigAnalyzer()
        result = analyzer.analyze(sample_nim_image)

        assert result.success
        assert result.report is not None

        high_impact = result.report.high_impact_entries
        # Should include entries like NIM_TENSOR_PARALLEL_SIZE if set

    def test_validate_config(self, sample_nim_image: NIMImage):
        """Test config validation."""
        analyzer = ConfigAnalyzer()

        # Valid config
        errors = analyzer.validate(
            sample_nim_image,
            env={"NIM_LOG_LEVEL": "DEBUG"},
        )
        assert len(errors) == 0

    def test_validate_invalid_value(self, sample_nim_image: NIMImage):
        """Test validation catches invalid values."""
        analyzer = ConfigAnalyzer()

        errors = analyzer.validate(
            sample_nim_image,
            env={"NIM_LOG_LEVEL": "INVALID"},
        )
        assert len(errors) > 0
        assert any("NIM_LOG_LEVEL" in e for e in errors)

    def test_load_env_file(self, sample_env_file: str):
        """Test loading env file."""
        analyzer = ConfigAnalyzer()
        env = analyzer._load_env_file(sample_env_file)

        assert env["NIM_SERVER_PORT"] == "8000"
        assert env["NIM_LOG_LEVEL"] == "DEBUG"
        assert env["NIM_MAX_BATCH_SIZE"] == "32"
        assert env["NIM_MODEL_NAME"] == "llama3-8b"  # Quoted value

    def test_effective_value(self, sample_nim_image: NIMImage):
        """Test effective value calculation."""
        analyzer = ConfigAnalyzer()
        result = analyzer.analyze(sample_nim_image)

        assert result.success
        assert result.report is not None

        for entry in result.report.entries:
            if entry.value is not None:
                assert entry.effective_value == entry.value
            else:
                assert entry.effective_value == entry.default_value
